@@ -50,7 +50,7 @@ import org.apache.catalina.util.CharsetMapper;
  * <p>
  *
  * @author Craig R. McClanahan
- * @version $Id: Context.java 1137629 2011-06-20 13:38:23Z markt $
+ *
  */
 
 public interface Context extends Container {
@@ -542,6 +542,101 @@ public interface Context extends Container {
     public void setWrapperClass(String wrapperClass);
 
 
+    /**
+     * Will the parsing of the web.xml file for this Context be performed by a
+     * namespace aware parser?
+     *
+     * @return true if namespace awareness is enabled.
+     */
+    public boolean getXmlNamespaceAware();
+
+
+    /**
+     * Controls whether the parsing of the web.xml file for this Context will be
+     * performed by a namespace aware parser.
+     *
+     * @param xmlNamespaceAware true to enable namespace awareness
+     */
+    public void setXmlNamespaceAware(boolean xmlNamespaceAware);
+     
+
+    /**
+     * Will the parsing of the web.xml file for this Context be performed by a
+     * validating parser?
+     *
+     * @return true if validation is enabled.
+     */
+    public boolean getXmlValidation();
+
+
+    /**
+     * Controls whether the parsing of the web.xml file for this Context will be
+     * performed by a validating parser.
+     *
+     * @param xmlValidation true to enable xml validation
+     */
+    public void setXmlValidation(boolean xmlValidation);
+
+
+    /**
+     * *.tld files are always parsed using a namespace aware parser.
+     *
+     * @return Always <code>true</code>
+     * 
+     * @deprecated This option will be removed in 8.0.x.
+     */
+    @Deprecated
+    public boolean getTldNamespaceAware();
+
+
+    /**
+     * *.tld files are always parsed using a namespace aware parser.
+     *
+     * @param tldNamespaceAware ignored
+     * 
+     * @deprecated This option will be removed in 8.0.x.
+     */
+    @Deprecated
+    public void setTldNamespaceAware(boolean tldNamespaceAware);
+
+
+    /**
+     * Will the parsing of web.xml, web-fragment.xml, *.tld, *.jspx, *.tagx and
+     * tagplugin.xml files for this Context block the use of external entities?
+     *
+     * @return true if access to external entities is blocked
+     */
+    public boolean getXmlBlockExternal();
+
+
+    /**
+     * Controls whether the parsing of web.xml, web-fragment.xml, *.tld, *.jspx,
+     * *.tagx and tagplugin.xml files for this Context will block the use of
+     * external entities.
+     *
+     * @param xmlBlockExternal true to block external entities
+     */
+    public void setXmlBlockExternal(boolean xmlBlockExternal);
+
+
+    /**
+     * Will the parsing of *.tld files for this Context be performed by a
+     * validating parser?
+     *
+     * @return true if validation is enabled.
+     */
+    public boolean getTldValidation();
+
+
+    /**
+     * Controls whether the parsing of *.tld files for this Context will be
+     * performed by a validating parser.
+     *
+     * @param tldValidation true to enable xml validation
+     */
+    public void setTldValidation(boolean tldValidation);
+
+
     // --------------------------------------------------------- Public Methods
 
 
@@ -1098,67 +1193,65 @@ public interface Context extends Container {
      */
     public void removeWrapperListener(String listener);
 
-
     /**
-     * Get the server.xml <context> attribute's xmlNamespaceAware.
-     * @return true if namespace awarenes is enabled.
+     * If enabled, requests for a web application context root will be
+     * redirected (adding a trailing slash) by the Mapper. This is more
+     * efficient but has the side effect of confirming that the context path is
+     * valid.
      *
+     * @param mapperContextRootRedirectEnabled Should the redirects be enabled?
      */
-    public boolean getXmlNamespaceAware();
-
+    public void setMapperContextRootRedirectEnabled(boolean mapperContextRootRedirectEnabled);
 
     /**
-     * Get the server.xml <context> attribute's xmlValidation.
-     * @return true if validation is enabled.
+     * Determines if requests for a web application context root will be
+     * redirected (adding a trailing slash) by the Mapper. This is more
+     * efficient but has the side effect of confirming that the context path is
+     * valid.
      *
+     * @return {@code true} if the Mapper level redirect is enabled for this
+     *         Context.
      */
-    public boolean getXmlValidation();
-
+    public boolean getMapperContextRootRedirectEnabled();
 
     /**
-     * Set the validation feature of the XML parser used when
-     * parsing xml instances.
-     * @param xmlValidation true to enable xml instance validation
+     * If enabled, requests for a directory will be redirected (adding a
+     * trailing slash) by the Mapper. This is more efficient but has the
+     * side effect of confirming that the directory is valid.
+     *
+     * @param mapperDirectoryRedirectEnabled Should the redirects be enabled?
      */
-    public void setXmlValidation(boolean xmlValidation);
-
-
-   /**
-     * Set the namespace aware feature of the XML parser used when
-     * parsing xml instances.
-     * @param xmlNamespaceAware true to enable namespace awareness
-     */
-    public void setXmlNamespaceAware(boolean xmlNamespaceAware);
-     
+    public void setMapperDirectoryRedirectEnabled(boolean mapperDirectoryRedirectEnabled);
 
     /**
-     * Set the validation feature of the XML parser used when
-     * parsing tlds files. 
-     * @param tldValidation true to enable xml instance validation
+     * Determines if requests for a directory will be redirected (adding a
+     * trailing slash) by the Mapper. This is more efficient but has the
+     * side effect of confirming that the directory is valid.
+     *
+     * @return {@code true} if the Mapper level redirect is enabled for this
+     *         Context.
      */
-    public void setTldValidation(boolean tldValidation);
-
+    public boolean getMapperDirectoryRedirectEnabled();
 
     /**
-     * Get the validation feature of the XML parser used when
-     * parsing tlds files. 
-     * @return true if validation is enabled.
+     * Are paths used in calls to obtain a request dispatcher expected to be
+     * encoded? This affects both how Tomcat handles calls to obtain a request
+     * dispatcher as well as how Tomcat generates paths used to obtain request
+     * dispatchers internally.
+     *
+     * @param dispatchersUseEncodedPaths {@code true} to use encoded paths,
+     *        otherwise {@code false}
      */
-    public boolean getTldValidation();
-
+    public void setDispatchersUseEncodedPaths(boolean dispatchersUseEncodedPaths);
 
     /**
-     * Get the server.xml &lt;host&gt; attribute's xmlNamespaceAware.
-     * @return true if namespace awarenes is enabled.
+     * Are paths used in calls to obtain a request dispatcher expected to be
+     * encoded? This applys to both how Tomcat handles calls to obtain a request
+     * dispatcher as well as how Tomcat generates paths used to obtain request
+     * dispatchers internally.
+     *
+     * @return {@code true} if encoded paths will be used, otherwise
+     *         {@code false}
      */
-    public boolean getTldNamespaceAware();
-
-
-    /**
-     * Set the namespace aware feature of the XML parser used when
-     * parsing xml instances.
-     * @param tldNamespaceAware true to enable namespace awareness
-     */
-    public void setTldNamespaceAware(boolean tldNamespaceAware);
+    public boolean getDispatchersUseEncodedPaths();
 }
-
