@@ -5,9 +5,9 @@
  * The ASF licenses this file to You under the Apache License, Version 2.0
  * (the "License"); you may not use this file except in compliance with
  * the License.  You may obtain a copy of the License at
- * 
+ *
  *      http://www.apache.org/licenses/LICENSE-2.0
- * 
+ *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -21,14 +21,19 @@ package org.apache.catalina.core;
 
 import java.io.File;
 import java.io.InputStream;
+import java.io.UnsupportedEncodingException;
 import java.net.MalformedURLException;
 import java.net.URL;
+import java.net.URLDecoder;
 import java.util.ArrayList;
+import java.util.Collections;
 import java.util.Enumeration;
+import java.util.HashSet;
 import java.util.Iterator;
 import java.util.Map;
 import java.util.Set;
 import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.ConcurrentMap;
 
 import javax.naming.Binding;
 import javax.naming.NamingException;
@@ -42,12 +47,12 @@ import javax.servlet.ServletContextAttributeListener;
 import org.apache.catalina.Context;
 import org.apache.catalina.Host;
 import org.apache.catalina.Wrapper;
-import org.apache.catalina.deploy.ApplicationParameter;
 import org.apache.catalina.util.Enumerator;
 import org.apache.catalina.util.RequestUtil;
 import org.apache.catalina.util.ResourceSet;
 import org.apache.catalina.util.ServerInfo;
 import org.apache.catalina.util.StringManager;
+import org.apache.catalina.util.URLEncoder;
 import org.apache.naming.resources.DirContextURLStreamHandler;
 import org.apache.naming.resources.Resource;
 import org.apache.tomcat.util.buf.CharChunk;
@@ -63,11 +68,9 @@ import org.apache.catalina.Globals;
  *
  * @author Craig R. McClanahan
  * @author Remy Maucherat
- * @version $Revision: 739532 $ $Date: 2009-01-31 10:52:13 +0100 (Sat, 31 Jan 2009) $
+ *
  */
-
-public class ApplicationContext
-    implements ServletContext {
+public class ApplicationContext implements ServletContext {
 
     // ----------------------------------------------------------- Constructors
 
@@ -97,9 +100,9 @@ public class ApplicationContext
     /**
      * List of read only attributes for this context.
      */
-    private Map readOnlyAttributes = new ConcurrentHashMap();
-
+    private final Map<String,String> readOnlyAttributes = new ConcurrentHashMap<String,String>();
 
+    
     /**
      * The Context instance with which we are associated.
      */
@@ -122,7 +125,7 @@ public class ApplicationContext
     /**
      * The merged context initialization parameters for this Context.
      */
-    private Map parameters = null;
+    private final ConcurrentMap<String,String> parameters = new ConcurrentHashMap<String,String>();
 
 
     /**
@@ -235,14 +238,14 @@ public class ApplicationContext
         }
     }
 
-    
+
     /**
      * Return the main path associated with this context.
      */
     public String getContextPath() {
         return context.getPath();
     }
-    
+
 
     /**
      * Return the value of the specified initialization parameter, or
@@ -251,10 +254,33 @@ public class ApplicationContext
      * @param name Name of the initialization parameter to retrieve
      */
     public String getInitParameter(final String name) {
+        // Special handling for XML settings as the context setting must
+        // always override anything that might have been set by an application.
+        if (Globals.JASPER_XML_VALIDATION_TLD_INIT_PARAM.equals(name) &&
+                context.getTldValidation()) {
+            return "true";
+        }
+        if (Globals.JASPER_XML_VALIDATION_INIT_PARAM.equals(name) &&
+                context.getXmlValidation()) {
+            return "true";
+        }
+        if (Globals.JASPER_XML_BLOCK_EXTERNAL_INIT_PARAM.equals(name)) {
+            if (!context.getXmlBlockExternal()) {
+                // System admin has explicitly changed the default
+                return "false";
+            }
+        }
+        return parameters.get(name);
+    }
 
-        mergeParameters();
-        return ((String) parameters.get(name));
 
+    public boolean setInitParameter(String name, String value) {
+        if (parameters.containsKey(name)) {
+            return false;
+        }
+
+        parameters.put(name, value);
+        return true;
     }
 
 
@@ -262,11 +288,21 @@ public class ApplicationContext
      * Return the names of the context's initialization parameters, or an
      * empty enumeration if the context has no initialization parameters.
      */
-    public Enumeration getInitParameterNames() {
-
-        mergeParameters();
-        return (new Enumerator(parameters.keySet()));
-
+    public Enumeration<String> getInitParameterNames() {
+        Set<String> names = new HashSet<String>();
+        names.addAll(parameters.keySet());
+        // Special handling for XML settings as these attributes will always be
+        // available if they have been set on the context
+        if (context.getTldValidation()) {
+            names.add(Globals.JASPER_XML_VALIDATION_TLD_INIT_PARAM);
+        }
+        if (context.getXmlValidation()) {
+            names.add(Globals.JASPER_XML_VALIDATION_INIT_PARAM);
+        }
+        if (!context.getXmlBlockExternal()) {
+            names.add(Globals.JASPER_XML_BLOCK_EXTERNAL_INIT_PARAM);
+        }
+        return Collections.enumeration(names);
     }
 
 
@@ -327,7 +363,7 @@ public class ApplicationContext
         Wrapper wrapper = (Wrapper) context.findChild(name);
         if (wrapper == null)
             return (null);
-        
+
         return new ApplicationDispatcher(wrapper, null, null, null, null, name);
 
     }
@@ -373,17 +409,38 @@ public class ApplicationContext
 
         // Get query string
         String queryString = null;
-        int pos = path.indexOf('?');
+        String normalizedPath = path;
+        int pos = normalizedPath.indexOf('?');
         if (pos >= 0) {
-            queryString = path.substring(pos + 1);
-            path = path.substring(0, pos);
+            queryString = normalizedPath.substring(pos + 1);
+            normalizedPath = normalizedPath.substring(0, pos);
         }
 
-        path = RequestUtil.normalize(path);
-        if (path == null)
+        normalizedPath = RequestUtil.normalize(normalizedPath);
+        if (normalizedPath == null)
             return (null);
 
-        pos = path.length(); 
+        if (getContext().getDispatchersUseEncodedPaths()) {
+            // Decode
+            String decodedPath;
+            try {
+                decodedPath = URLDecoder.decode(normalizedPath, "UTF-8");
+            } catch (UnsupportedEncodingException e) {
+                // Impossible
+                return null;
+            }
+
+            // Security check to catch attempts to encode /../ sequences
+            normalizedPath = RequestUtil.normalize(decodedPath);
+            if (!decodedPath.equals(normalizedPath)) {
+                getContext().getLogger().warn(
+                        sm.getString("applicationContext.illegalDispatchPath", path),
+                        new IllegalArgumentException());
+                return null;
+            }
+        }
+
+        pos = normalizedPath.length();
 
         // Use the thread local URI and mapping data
         DispatchData dd = dispatchData.get();
@@ -406,11 +463,11 @@ public class ApplicationContext
              * Ignore any trailing path params (separated by ';') for mapping
              * purposes
              */
-            int semicolon = path.indexOf(';');
+            int semicolon = normalizedPath.indexOf(';');
             if (pos >= 0 && semicolon > pos) {
                 semicolon = -1;
             }
-            uriCC.append(path, 0, semicolon > 0 ? semicolon : pos);
+            uriCC.append(normalizedPath, 0, semicolon > 0 ? semicolon : pos);
             context.getMapper().map(uriMB, mappingData);
             if (mappingData.wrapper == null) {
                 return (null);
@@ -421,7 +478,7 @@ public class ApplicationContext
              * RequestDispatcher's requestURI
              */
             if (semicolon > 0) {
-                uriCC.append(path, semicolon, pos - semicolon);
+                uriCC.append(normalizedPath, semicolon, pos - semicolon);
             }
         } catch (Exception e) {
             // Should never happen
@@ -434,12 +491,12 @@ public class ApplicationContext
         String pathInfo = mappingData.pathInfo.toString();
 
         mappingData.recycle();
-        
-        // Construct a RequestDispatcher to process this request
-        return new ApplicationDispatcher
-            (wrapper, uriCC.toString(), wrapperPath, pathInfo, 
-             queryString, null);
 
+        String encodedUri = URLEncoder.DEFAULT.encode(uriCC.toString(), "UTF-8");
+
+        // Construct a RequestDispatcher to process this request
+        return new ApplicationDispatcher(wrapper, encodedUri, wrapperPath, pathInfo,
+             queryString, null);
     }
 
 
@@ -463,7 +520,7 @@ public class ApplicationContext
         if (!path.startsWith("/") && Globals.STRICT_SERVLET_COMPLIANCE)
             throw new MalformedURLException(sm.getString("applicationContext.requestDispatcher.iae", path));
 
-        
+
         path = RequestUtil.normalize(path);
         if (path == null)
             return (null);
@@ -492,8 +549,12 @@ public class ApplicationContext
                     return new URL
                         ("jndi", "", 0, getJNDIUri(hostName, fullPath),
                          new DirContextURLStreamHandler(resources));
-                } catch (Exception e) {
+                } catch (NamingException e) {
                     // Ignore
+                } catch (Exception e) {
+                    // Unexpected
+                    log(sm.getString("applicationContext.lookup.error", path,
+                            getContextPath()), e);
                 }
             }
         }
@@ -529,7 +590,12 @@ public class ApplicationContext
                 Object resource = resources.lookup(path);
                 if (resource instanceof Resource)
                     return (((Resource) resource).streamContent());
+            } catch (NamingException e) {
+                // Ignore
             } catch (Exception e) {
+                // Unexpected
+                log(sm.getString("applicationContext.lookup.error", path,
+                        getContextPath()), e);
             }
         }
         return (null);
@@ -656,7 +722,7 @@ public class ApplicationContext
      *  <code>log(String, Throwable)</code> instead
      */
     public void log(Exception exception, String message) {
-        
+
         context.getLogger().error(message, exception);
 
     }
@@ -669,7 +735,7 @@ public class ApplicationContext
      * @param throwable Exception to be reported
      */
     public void log(String message, Throwable throwable) {
-        
+
         context.getLogger().error(message, throwable);
 
     }
@@ -810,7 +876,7 @@ public class ApplicationContext
     protected StandardContext getContext() {
         return this.context;
     }
-    
+
     protected Map getReadonlyAttributes() {
         return this.readOnlyAttributes;
     }
@@ -833,10 +899,10 @@ public class ApplicationContext
             String key = (String) keys.next();
             removeAttribute(key);
         }
-        
+
     }
-    
-    
+
+
     /**
      * Return the facade associated with this ApplicationContext.
      */
@@ -858,38 +924,6 @@ public class ApplicationContext
     }
 
 
-    // -------------------------------------------------------- Private Methods
-
-
-    /**
-     * Merge the context initialization parameters specified in the application
-     * deployment descriptor with the application parameters described in the
-     * server configuration, respecting the <code>override</code> property of
-     * the application parameters appropriately.
-     */
-    private void mergeParameters() {
-
-        if (parameters != null)
-            return;
-        Map results = new ConcurrentHashMap();
-        String names[] = context.findParameters();
-        for (int i = 0; i < names.length; i++)
-            results.put(names[i], context.findParameter(names[i]));
-        ApplicationParameter params[] =
-            context.findApplicationParameters();
-        for (int i = 0; i < params.length; i++) {
-            if (params[i].getOverride()) {
-                if (results.get(params[i].getName()) == null)
-                    results.put(params[i].getName(), params[i].getValue());
-            } else {
-                results.put(params[i].getName(), params[i].getValue());
-            }
-        }
-        parameters = results;
-
-    }
-
-
     /**
      * List resource paths (recursively), and store all of them in the given
      * Set.
