@@ -5,9 +5,9 @@
  * The ASF licenses this file to You under the Apache License, Version 2.0
  * (the "License"); you may not use this file except in compliance with
  * the License.  You may obtain a copy of the License at
- * 
+ *
  *      http://www.apache.org/licenses/LICENSE-2.0
- * 
+ *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -37,6 +37,7 @@ import javax.management.ObjectName;
 import javax.naming.directory.DirContext;
 import javax.servlet.ServletException;
 
+import org.apache.catalina.AccessLog;
 import org.apache.catalina.Cluster;
 import org.apache.catalina.Container;
 import org.apache.catalina.ContainerEvent;
@@ -52,6 +53,7 @@ import org.apache.catalina.Realm;
 import org.apache.catalina.Valve;
 import org.apache.catalina.connector.Request;
 import org.apache.catalina.connector.Response;
+import org.apache.catalina.session.ManagerBase;
 import org.apache.catalina.util.LifecycleSupport;
 import org.apache.catalina.util.StringManager;
 import org.apache.juli.logging.Log;
@@ -192,7 +194,7 @@ public abstract class ContainerBase
      * Associated logger name.
      */
     protected String logName = null;
-    
+
 
     /**
      * The Manager implementation with which this Container is associated.
@@ -205,7 +207,7 @@ public abstract class ContainerBase
      */
     protected Cluster cluster = null;
 
-    
+
     /**
      * The human-readable name of this Container.
      */
@@ -279,16 +281,23 @@ public abstract class ContainerBase
     private boolean threadDone = false;
 
 
+    /**
+     * The access log to use for requests normally handled by this container
+     * that have been handled earlier in the processing chain.
+     */
+    protected volatile AccessLog accessLog = null;
+    private volatile boolean accessLogScanComplete = false;
+
     // ------------------------------------------------------------- Properties
 
 
     /**
      * Get the delay between the invocation of the backgroundProcess method on
      * this container and its children. Child containers will not be invoked
-     * if their delay value is not negative (which would mean they are using 
-     * their own thread). Setting this to a positive value will cause 
-     * a thread to be spawn. After waiting the specified amount of time, 
-     * the thread will invoke the executePeriodic method on this container 
+     * if their delay value is not negative (which would mean they are using
+     * their own thread). Setting this to a positive value will cause
+     * a thread to be spawn. After waiting the specified amount of time,
+     * the thread will invoke the executePeriodic method on this container
      * and all its children.
      */
     public int getBackgroundProcessorDelay() {
@@ -299,8 +308,8 @@ public abstract class ContainerBase
     /**
      * Set the delay between the invocation of the execute method on this
      * container and its children.
-     * 
-     * @param delay The delay in seconds between the invocation of 
+     *
+     * @param delay The delay in seconds between the invocation of
      *              backgroundProcess methods
      */
     public void setBackgroundProcessorDelay(int delay) {
@@ -420,20 +429,21 @@ public abstract class ContainerBase
         this.manager = manager;
 
         // Stop the old component if necessary
-        if (started && (oldManager != null) &&
-            (oldManager instanceof Lifecycle)) {
+        if (oldManager instanceof Lifecycle) {
             try {
                 ((Lifecycle) oldManager).stop();
+                if (oldManager instanceof ManagerBase) {
+                    ((ManagerBase) oldManager).destroy();
+                }
             } catch (LifecycleException e) {
-                log.error("ContainerBase.setManager: stop: ", e);
+                log.error("ContainerBase.setManager: stop-destroy: ", e);
             }
         }
 
         // Start the new component if necessary
         if (manager != null)
             manager.setContainer(this);
-        if (started && (manager != null) &&
-            (manager instanceof Lifecycle)) {
+        if (started && manager instanceof Lifecycle) {
             try {
                 ((Lifecycle) manager).start();
             } catch (LifecycleException e) {
@@ -443,7 +453,6 @@ public abstract class ContainerBase
 
         // Report this property change to interested listeners
         support.firePropertyChange("manager", oldManager, this.manager);
-
     }
 
 
@@ -724,8 +733,8 @@ public abstract class ContainerBase
      */
     public synchronized void setResources(DirContext resources) {
         // Called from StandardContext.setResources()
-        //              <- StandardContext.start() 
-        //              <- ContainerBase.addChildInternal() 
+        //              <- StandardContext.start()
+        //              <- ContainerBase.addChildInternal()
 
         // Change components if necessary
         DirContext oldResources = this.resources;
@@ -783,27 +792,30 @@ public abstract class ContainerBase
                                                    "' is not unique");
             child.setParent(this);  // May throw IAE
             children.put(child.getName(), child);
+        }
 
-            // Start child
-            if (started && startChildren && (child instanceof Lifecycle)) {
-                boolean success = false;
-                try {
-                    ((Lifecycle) child).start();
-                    success = true;
-                } catch (LifecycleException e) {
-                    log.error("ContainerBase.addChild: start: ", e);
-                    throw new IllegalStateException
-                        ("ContainerBase.addChild: start: " + e);
-                } finally {
-                    if (!success) {
+        // Start child
+        // Don't do this inside sync block - start can be a slow process and
+        // locking the children object can cause problems elsewhere
+        if (started && startChildren && (child instanceof Lifecycle)) {
+            boolean success = false;
+            try {
+                ((Lifecycle) child).start();
+                success = true;
+            } catch (LifecycleException e) {
+                log.error("ContainerBase.addChild: start: ", e);
+                throw new IllegalStateException
+                    ("ContainerBase.addChild: start: " + e);
+            } finally {
+                if (!success) {
+                    synchronized(children) {
                         children.remove(child.getName());
                     }
                 }
             }
-
-            fireContainerEvent(ADD_CHILD_EVENT, child);
         }
 
+        fireContainerEvent(ADD_CHILD_EVENT, child);
     }
 
 
@@ -843,7 +855,7 @@ public abstract class ContainerBase
 
         if (name == null)
             return (null);
-        synchronized (children) {       // Required by post-start changes
+        synchronized (children) {
             return ((Container) children.get(name));
         }
 
@@ -872,7 +884,7 @@ public abstract class ContainerBase
     public ContainerListener[] findContainerListeners() {
 
         synchronized (listeners) {
-            ContainerListener[] results = 
+            ContainerListener[] results =
                 new ContainerListener[listeners.size()];
             return ((ContainerListener[]) listeners.toArray(results));
         }
@@ -911,12 +923,10 @@ public abstract class ContainerBase
      */
     public void removeChild(Container child) {
 
-        synchronized(children) {
-            if (children.get(child.getName()) == null)
-                return;
-            children.remove(child.getName());
+        if (child == null) {
+            return;
         }
-        
+
         if (started && (child instanceof Lifecycle)) {
             try {
                 if( child instanceof ContainerBase ) {
@@ -930,9 +940,15 @@ public abstract class ContainerBase
                 log.error("ContainerBase.removeChild: stop: ", e);
             }
         }
-        
+
+        synchronized(children) {
+            if (children.get(child.getName()) == null)
+                return;
+            children.remove(child.getName());
+        }
+
         fireContainerEvent(REMOVE_CHILD_EVENT, child);
-        
+
         // child.setParent(null);
 
     }
@@ -980,7 +996,7 @@ public abstract class ContainerBase
 
 
     /**
-     * Get the lifecycle listeners associated with this lifecycle. If this 
+     * Get the lifecycle listeners associated with this lifecycle. If this
      * Lifecycle has no listeners registered, a zero-length array is returned.
      */
     public LifecycleListener[] findLifecycleListeners() {
@@ -1016,7 +1032,7 @@ public abstract class ContainerBase
                 log.info(sm.getString("containerBase.alreadyStarted", logName()));
             return;
         }
-        
+
         // Notify our interested LifecycleListeners
         lifecycle.fireLifecycleEvent(BEFORE_START_EVENT, null);
 
@@ -1129,15 +1145,15 @@ public abstract class ContainerBase
     }
 
     /** Init method, part of the MBean lifecycle.
-     *  If the container was added via JMX, it'll register itself with the 
+     *  If the container was added via JMX, it'll register itself with the
      * parent, using the ObjectName conventions to locate the parent.
-     * 
+     *
      *  If the container was added directly and it doesn't have an ObjectName,
-     * it'll create a name and register itself with the JMX console. On destroy(), 
+     * it'll create a name and register itself with the JMX console. On destroy(),
      * the object will unregister.
-     * 
+     *
      * @throws Exception
-     */ 
+     */
     public void init() throws Exception {
 
         if( this.getParent() == null ) {
@@ -1145,8 +1161,8 @@ public abstract class ContainerBase
             ObjectName parentName=getParentName();
 
             //log.info("Register " + parentName );
-            if( parentName != null && 
-                    mserver.isRegistered(parentName)) 
+            if( parentName != null &&
+                    mserver.isRegistered(parentName))
             {
                 mserver.invoke(parentName, "addChild", new Object[] { this },
                         new String[] {"org.apache.catalina.Container"});
@@ -1154,11 +1170,11 @@ public abstract class ContainerBase
         }
         initialized=true;
     }
-    
+
     public ObjectName getParentName() throws MalformedObjectNameException {
         return null;
     }
-    
+
     public void destroy() throws Exception {
         if( started ) {
             stop();
@@ -1188,7 +1204,47 @@ public abstract class ContainerBase
         for (int i = 0; i < children.length; i++) {
             removeChild(children[i]);
         }
-                
+
+    }
+
+
+    /**
+     * Check this container for an access log and if none is found, look to the
+     * parent. If there is no parent and still none is found, use the NoOp
+     * access log.
+     */
+    public void logAccess(Request request, Response response, long time,
+            boolean useDefault) {
+
+        boolean logged = false;
+
+        if (getAccessLog() != null) {
+            getAccessLog().log(request, response, time);
+            logged = true;
+        }
+
+        if (getParent() != null) {
+            // No need to use default logger once request/response has been logged
+            // once
+            getParent().logAccess(request, response, time, (useDefault && !logged));
+        }
+    }
+
+    public AccessLog getAccessLog() {
+
+        if (accessLogScanComplete) {
+            return accessLog;
+        }
+
+        Valve valves[] = getPipeline().getValves();
+        for (Valve valve : valves) {
+            if (valve instanceof AccessLog) {
+                accessLog = (AccessLog) valve;
+                break;
+            }
+        }
+        accessLogScanComplete = true;
+        return accessLog;
     }
 
     // ------------------------------------------------------- Pipeline Methods
@@ -1221,7 +1277,7 @@ public abstract class ContainerBase
     public ObjectName[] getValveObjectNames() {
         return ((StandardPipeline)pipeline).getValveObjectNames();
     }
-    
+
     /**
      * <p>Return the Valve instance that has been distinguished as the basic
      * Valve for this Pipeline (if any).
@@ -1293,7 +1349,7 @@ public abstract class ContainerBase
      * throwables will be caught and logged.
      */
     public void backgroundProcess() {
-        
+
         if (!started)
             return;
 
@@ -1301,28 +1357,28 @@ public abstract class ContainerBase
             try {
                 cluster.backgroundProcess();
             } catch (Exception e) {
-                log.warn(sm.getString("containerBase.backgroundProcess.cluster", cluster), e);                
+                log.warn(sm.getString("containerBase.backgroundProcess.cluster", cluster), e);
             }
         }
         if (loader != null) {
             try {
                 loader.backgroundProcess();
             } catch (Exception e) {
-                log.warn(sm.getString("containerBase.backgroundProcess.loader", loader), e);                
+                log.warn(sm.getString("containerBase.backgroundProcess.loader", loader), e);
             }
         }
         if (manager != null) {
             try {
                 manager.backgroundProcess();
             } catch (Exception e) {
-                log.warn(sm.getString("containerBase.backgroundProcess.manager", manager), e);                
+                log.warn(sm.getString("containerBase.backgroundProcess.manager", manager), e);
             }
         }
         if (realm != null) {
             try {
                 realm.backgroundProcess();
             } catch (Exception e) {
-                log.warn(sm.getString("containerBase.backgroundProcess.realm", realm), e);                
+                log.warn(sm.getString("containerBase.backgroundProcess.realm", realm), e);
             }
         }
         Valve current = pipeline.getFirst();
@@ -1330,7 +1386,7 @@ public abstract class ContainerBase
             try {
                 current.backgroundProcess();
             } catch (Exception e) {
-                log.warn(sm.getString("containerBase.backgroundProcess.valve", current), e);                
+                log.warn(sm.getString("containerBase.backgroundProcess.valve", current), e);
             }
             current = current.getNext();
         }
@@ -1379,16 +1435,16 @@ public abstract class ContainerBase
             if ((name == null) || (name.equals(""))) {
                 name = "/";
             }
-            loggerName = "[" + name + "]" 
+            loggerName = "[" + name + "]"
                 + ((loggerName != null) ? ("." + loggerName) : "");
             current = current.getParent();
         }
         logName = ContainerBase.class.getName() + "." + loggerName;
         return logName;
-        
+
     }
 
-    
+
     // -------------------- JMX and Registration  --------------------
     protected String type;
     protected String domain;
@@ -1400,7 +1456,7 @@ public abstract class ContainerBase
     public ObjectName getJmxName() {
         return oname;
     }
-    
+
     public String getObjectName() {
         if (oname != null) {
             return oname.toString();
@@ -1416,7 +1472,7 @@ public abstract class ContainerBase
             }
             if( parent instanceof StandardEngine ) {
                 domain=((StandardEngine)parent).getDomain();
-            } 
+            }
         }
         return domain;
     }
@@ -1424,7 +1480,7 @@ public abstract class ContainerBase
     public void setDomain(String domain) {
         this.domain=domain;
     }
-    
+
     public String getType() {
         return type;
     }
@@ -1495,9 +1551,9 @@ public abstract class ContainerBase
         Container context=null;
         Container host=null;
         Container servlet=null;
-        
+
         StringBuffer suffix=new StringBuffer();
-        
+
         if( container instanceof StandardHost ) {
             host=container;
         } else if( container instanceof StandardContext ) {
@@ -1511,7 +1567,7 @@ public abstract class ContainerBase
         if( context!=null ) {
             String path=((StandardContext)context).getPath();
             suffix.append(",path=").append((path.equals("")) ? "/" : path);
-        } 
+        }
         if( host!=null ) suffix.append(",host=").append( host.getName() );
         if( servlet != null ) {
             String name=container.getName();
@@ -1568,7 +1624,7 @@ public abstract class ContainerBase
 
 
     /**
-     * Private thread class to invoke the backgroundProcess method 
+     * Private thread class to invoke the backgroundProcess method
      * of this container and its children after a fixed delay.
      */
     protected class ContainerBackgroundProcessor implements Runnable {
@@ -1582,7 +1638,7 @@ public abstract class ContainerBase
                 }
                 if (!threadDone) {
                     Container parent = (Container) getMappingObject();
-                    ClassLoader cl = 
+                    ClassLoader cl =
                         Thread.currentThread().getContextClassLoader();
                     if (parent.getLoader() != null) {
                         cl = parent.getLoader().getClassLoader();
@@ -1614,5 +1670,10 @@ public abstract class ContainerBase
 
     }
 
+    protected static final class NoopAccessLog implements AccessLog {
 
+        public void log(Request request, Response response, long time) {
+            // NOOP
+        }
+    }
 }
