@@ -41,6 +41,7 @@ import org.apache.catalina.ha.CatalinaCluster;
 import org.apache.catalina.ha.ClusterMessage;
 import org.apache.catalina.ha.tcp.ReplicationValve;
 import org.apache.catalina.session.ManagerBase;
+import org.apache.catalina.session.TooManyActiveSessionsException;
 import org.apache.catalina.tribes.Member;
 import org.apache.catalina.tribes.io.ReplicationStream;
 import org.apache.catalina.util.LifecycleSupport;
@@ -63,7 +64,7 @@ import org.apache.catalina.ha.ClusterManager;
  * @author Craig R. McClanahan
  * @author Jean-Francois Arcand
  * @author Peter Rossbach
- * @version $Id: DeltaManager.java 1152894 2011-08-01 19:10:01Z markt $
+ *
  */
 
 public class DeltaManager extends ClusterManagerBase{
@@ -115,6 +116,7 @@ public class DeltaManager extends ClusterManagerBase{
     private boolean notifySessionListenersOnReplication = true;
     private boolean notifyContainerListenersOnReplication  = true;
     private volatile boolean stateTransfered = false ;
+    private volatile boolean noContextManagerReceived = false ;
     private int stateTransferTimeout = 60;
     private boolean sendAllSessions = true;
     private boolean sendClusterDomainOnly = true ;
@@ -142,6 +144,7 @@ public class DeltaManager extends ClusterManagerBase{
     private long counterReceive_EVT_SESSION_DELTA = 0;
     private int counterReceive_EVT_ALL_SESSION_TRANSFERCOMPLETE = 0 ;
     private long counterReceive_EVT_CHANGE_SESSION_ID = 0 ;
+    private long counterReceive_EVT_ALL_SESSION_NOCONTEXTMANAGER = 0 ;
     private long counterSend_EVT_GET_ALL_SESSIONS = 0 ;
     private long counterSend_EVT_ALL_SESSION_DATA = 0 ;
     private long counterSend_EVT_SESSION_CREATED = 0;
@@ -291,7 +294,14 @@ public class DeltaManager extends ClusterManagerBase{
     public long getCounterReceive_EVT_CHANGE_SESSION_ID() {
         return counterReceive_EVT_CHANGE_SESSION_ID;
     }
-    
+
+    /**
+     * @return Returns the counterReceive_EVT_ALL_SESSION_NOCONTEXTMANAGER.
+     */
+    public long getCounterReceive_EVT_ALL_SESSION_NOCONTEXTMANAGER() {
+        return counterReceive_EVT_ALL_SESSION_NOCONTEXTMANAGER;
+    }
+
     /**
      * @return Returns the processingTime.
      */
@@ -358,7 +368,15 @@ public class DeltaManager extends ClusterManagerBase{
     public void setStateTransfered(boolean stateTransfered) {
         this.stateTransfered = stateTransfered;
     }
-    
+
+    public boolean isNoContextManagerReceived() {
+        return noContextManagerReceived;
+    }
+
+    public void setNoContextManagerReceived(boolean noContextManagerReceived) {
+        this.noContextManagerReceived = noContextManagerReceived;
+    }
+
     /**
      * @return Returns the sendAllSessionsWaitTime in msec
      */
@@ -528,7 +546,6 @@ public class DeltaManager extends ClusterManagerBase{
 
         // Register with the new Container (if any)
         if ((this.container != null) && (this.container instanceof Context)) {
-            setMaxInactiveInterval(((Context) this.container).getSessionTimeout() * 60);
             ((Context) this.container).addPropertyChangeListener(this);
         }
 
@@ -569,7 +586,9 @@ public class DeltaManager extends ClusterManagerBase{
     public Session createSession(String sessionId, boolean distribute) {
         if ((maxActiveSessions >= 0) && (sessions.size() >= maxActiveSessions)) {
             rejectedSessions++;
-            throw new IllegalStateException(sm.getString("deltaManager.createSession.ise"));
+            throw new TooManyActiveSessionsException(
+                    sm.getString("deltaManager.createSession.ise"),
+                    maxActiveSessions);
         }
         DeltaSession session = (DeltaSession) super.createSession(sessionId) ;
         if (distribute) {
@@ -898,25 +917,23 @@ public class DeltaManager extends ClusterManagerBase{
             // stop remove cluster binding
             //wow, how many nested levels of if statements can we have ;)
             if(cluster == null) {
-                Container context = getContainer() ;
-                if(context != null && context instanceof Context) {
-                     Container host = context.getParent() ;
-                     if(host != null && host instanceof Host) {
-                         cluster = host.getCluster();
-                         if(cluster != null && cluster instanceof CatalinaCluster) {
-                             setCluster((CatalinaCluster) cluster) ;
-                         } else {
-                             Container engine = host.getParent() ;
-                             if(engine != null && engine instanceof Engine) {
-                                 cluster = engine.getCluster();
-                                 if(cluster != null && cluster instanceof CatalinaCluster) {
-                                     setCluster((CatalinaCluster) cluster) ;
-                                 }
-                             } else {
-                                     cluster = null ;
-                             }
-                         }
-                     }
+                Container context = getContainer();
+                Container host = context.getParent() ;
+                if (host instanceof Host) {
+                    cluster = host.getCluster();
+                    if (cluster instanceof CatalinaCluster) {
+                        setCluster((CatalinaCluster) cluster) ;
+                    } else {
+                        Container engine = host.getParent() ;
+                        if (engine instanceof Engine) {
+                            cluster = engine.getCluster();
+                            if (cluster instanceof CatalinaCluster) {
+                                setCluster((CatalinaCluster) cluster) ;
+                            }
+                        } else {
+                            cluster = null ;
+                        }
+                    }
                 }
             }
             if (cluster == null) {
@@ -1055,7 +1072,7 @@ public class DeltaManager extends ClusterManagerBase{
                 }
                 reqNow = System.currentTimeMillis();
                 isTimeout = ((reqNow - reqStart) > (1000 * getStateTransferTimeout()));
-            } while ((!getStateTransfered()) && (!isTimeout));
+            } while ((!getStateTransfered()) && (!isTimeout) && (!isNoContextManagerReceived()));
         } else {
             if(getStateTransferTimeout() == -1) {
                 // wait that state is transfered
@@ -1064,16 +1081,19 @@ public class DeltaManager extends ClusterManagerBase{
                         Thread.sleep(100);
                     } catch (Exception sleep) {
                     }
-                } while ((!getStateTransfered()));
+                } while ((!getStateTransfered())&& (!isNoContextManagerReceived()));
                 reqNow = System.currentTimeMillis();
             }
         }
-        if (isTimeout || (!getStateTransfered())) {
+        if (isTimeout) {
             counterNoStateTransfered++ ;
-            log.error(sm.getString("deltaManager.noSessionState",getName(),new Date(beforeSendTime),new Long(reqNow - beforeSendTime)));
+            log.error(sm.getString("deltaManager.noSessionState",getName(),new Date(beforeSendTime),Long.valueOf(reqNow - beforeSendTime)));
+        } else if (isNoContextManagerReceived()) {
+            if (log.isWarnEnabled())
+                log.warn(sm.getString("deltaManager.noContextManager",getName(),new Date(beforeSendTime),Long.valueOf(reqNow - beforeSendTime)));
         } else {
             if (log.isInfoEnabled())
-                log.info(sm.getString("deltaManager.sessionReceived",getName(), new Date(beforeSendTime), new Long(reqNow - beforeSendTime)));
+                log.info(sm.getString("deltaManager.sessionReceived",getName(), new Date(beforeSendTime), Long.valueOf(reqNow - beforeSendTime)));
         }
     }
 
@@ -1130,19 +1150,7 @@ public class DeltaManager extends ClusterManagerBase{
      *            The property change event that has occurred
      */
     public void propertyChange(PropertyChangeEvent event) {
-
-        // Validate the source of this event
-        if (!(event.getSource() instanceof Context))
-            return;
-        // Process a relevant property change
-        if (event.getPropertyName().equals("sessionTimeout")) {
-            try {
-                setMaxInactiveInterval(((Integer) event.getNewValue()).intValue() * 60);
-            } catch (NumberFormatException e) {
-                log.error(sm.getString("deltaManager.sessionTimeout", event.getNewValue()));
-            }
-        }
-
+        // no-op
     }
 
     // -------------------------------------------------------- Replication
@@ -1361,7 +1369,7 @@ public class DeltaManager extends ClusterManagerBase{
     }
 
     /**
-     * Exipre all find sessions.
+     * Expire all find sessions.
      */
     public void expireAllLocalSessions()
     {
@@ -1385,7 +1393,7 @@ public class DeltaManager extends ClusterManagerBase{
             }//end if
         }//for
         long timeEnd = System.currentTimeMillis();
-        if(log.isDebugEnabled()) log.debug("End expire sessions " + getName() + " exipre processingTime " + (timeEnd - timeNow) + " expired direct sessions: " + expireDirect + " expired direct sessions: " + expireIndirect);
+        if(log.isDebugEnabled()) log.debug("End expire sessions " + getName() + " expire processingTime " + (timeEnd - timeNow) + " expired direct sessions: " + expireDirect + " expired direct sessions: " + expireIndirect);
       
     }
     
@@ -1473,6 +1481,10 @@ public class DeltaManager extends ClusterManagerBase{
                     handleCHANGE_SESSION_ID(msg,sender);
                     break;
                  }
+                case SessionMessage.EVT_ALL_SESSION_NOCONTEXTMANAGER: {
+                    handleALL_SESSION_NOCONTEXTMANAGER(msg,sender);
+                    break;
+                 }
                 default: {
                     //we didn't recognize the message type, do nothing
                     break;
@@ -1570,7 +1582,7 @@ public class DeltaManager extends ClusterManagerBase{
         session.setPrimarySession(false);
         session.setCreationTime(msg.getTimestamp());
         // use container maxInactiveInterval so that session will expire correctly in case of primary transfer
-        session.setMaxInactiveInterval(getMaxInactiveInterval(), false);
+        session.setMaxInactiveInterval(((Context) getContainer()).getSessionTimeout() * 60, false);
         session.access();
         session.setId(msg.getSessionID(), notifySessionListenersOnReplication);
         session.resetDeltaRequest();
@@ -1660,6 +1672,18 @@ public class DeltaManager extends ClusterManagerBase{
         }
     }
 
+    /**
+     * handle receive no context manager.
+     * @param msg
+     * @param sender
+     */
+    protected void handleALL_SESSION_NOCONTEXTMANAGER(SessionMessage msg, Member sender) {
+        counterReceive_EVT_ALL_SESSION_NOCONTEXTMANAGER++ ;
+        if (log.isDebugEnabled()) 
+            log.debug(sm.getString("deltaManager.receiveMessage.noContextManager",getName(), sender.getHost(), Integer.valueOf(sender.getPort())));
+        noContextManagerReceived = true ;
+    }
+
     /**
      * send a block of session to sender
      * @param sender
@@ -1695,6 +1719,7 @@ public class DeltaManager extends ClusterManagerBase{
         result.receiverQueue = receiverQueue ;
         result.stateTimestampDrop = stateTimestampDrop ;
         result.stateTransferCreateSendTime = stateTransferCreateSendTime; 
+        result.setSessionAttributeFilter(getSessionAttributeFilter());
         return result;
     }
 }
