@@ -45,7 +45,7 @@ import org.apache.catalina.util.StringManager;
 * @author Craig R. McClanahan
 * @author Carson McDonald
 * @author Ignacio Ortega
-* @version $Revision: 747840 $
+*
 */
 
 public class DataSourceRealm
@@ -251,6 +251,18 @@ public class DataSourceRealm
       this.userTable = userTable;
     }
 
+    /**
+     * Return descriptive information about this Realm implementation and
+     * the corresponding version number, in the format
+     * <code>&lt;description&gt;/&lt;version&gt;</code>.
+     */
+    @Override
+    public String getInfo() {
+
+        return info;
+
+    }
+
 
     // --------------------------------------------------------- Public Methods
 
@@ -268,6 +280,7 @@ public class DataSourceRealm
      * @param credentials Password or other credentials to use in
      *  authenticating this username
      */
+    @Override
     public Principal authenticate(String username, String credentials) {
     	
     	// No user or no credentials
@@ -290,13 +303,6 @@ public class DataSourceRealm
             // Acquire a Principal object for this user
             return authenticate(dbConnection, username, credentials);
             
-        } catch (SQLException e) {
-            // Log the problem for posterity
-            containerLog.error(sm.getString("dataSourceRealm.exception"), e);
-
-            // Return "not authenticated" for this request
-            return (null);
-
         } finally {
         	close(dbConnection);
         }
@@ -321,7 +327,7 @@ public class DataSourceRealm
      */
     protected Principal authenticate(Connection dbConnection,
                                                String username,
-                                               String credentials) throws SQLException{
+                                               String credentials) {
 
         String dbCredentials = getPassword(dbConnection, username);
 
@@ -530,7 +536,13 @@ public class DataSourceRealm
      */
     protected ArrayList<String> getRoles(Connection dbConnection,
                                      String username) {
-    	
+
+        if (allRolesMode != AllRolesMode.STRICT_MODE && !isRoleStoreDefined()) {
+            // Using an authentication only configuration and no role store has
+            // been defined so don't spend cycles looking
+            return null;
+        }
+
         ResultSet rs = null;
         PreparedStatement stmt = null;
         ArrayList<String> list = null;
@@ -610,8 +622,13 @@ public class DataSourceRealm
 
     }
 
-    // ------------------------------------------------------ Lifecycle Methods
 
+    private boolean isRoleStoreDefined() {
+        return userRoleTable != null || roleNameCol != null;
+    }
+
+
+    // ------------------------------------------------------ Lifecycle Methods
 
     /**
      *
