@@ -5,9 +5,9 @@
  * The ASF licenses this file to You under the Apache License, Version 2.0
  * (the "License"); you may not use this file except in compliance with
  * the License.  You may obtain a copy of the License at
- * 
+ *
  *      http://www.apache.org/licenses/LICENSE-2.0
- * 
+ *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -24,8 +24,14 @@ import java.security.Principal;
 import java.text.MessageFormat;
 import java.util.ArrayList;
 import java.util.Arrays;
+import java.util.Collections;
+import java.util.HashMap;
 import java.util.Hashtable;
+import java.util.Iterator;
 import java.util.List;
+import java.util.Map;
+import java.util.Set;
+import java.util.Map.Entry;
 
 import javax.naming.Context;
 import javax.naming.CommunicationException;
@@ -37,6 +43,7 @@ import javax.naming.NamingException;
 import javax.naming.NameParser;
 import javax.naming.Name;
 import javax.naming.AuthenticationException;
+import javax.naming.PartialResultException;
 import javax.naming.ServiceUnavailableException;
 import javax.naming.directory.Attribute;
 import javax.naming.directory.Attributes;
@@ -134,6 +141,15 @@ import org.apache.tomcat.util.buf.CharChunk;
  * in the user's element whose name is configured by the
  * <code>userRoleName</code> property.</li>
  *
+ * <li>A default role can be assigned to each user that was successfully
+ * authenticated by setting the <code>commonRole</code> property to the
+ * name of this role. The role doesn't have to exist in the directory.</li>
+ *
+ * <li>If the directory server contains nested roles, you can search for them
+ * by setting <code>roleNested</code> to <code>true</code>.
+ * The default value is <code>false</code>, so role searches will not find
+ * nested roles.</li>
+ *
  * <li>Note that the standard <code>&lt;security-role-ref&gt;</code> element in
  *     the web application deployment descriptor allows applications to refer
  *     to roles programmatically by names other than those used in the
@@ -147,13 +163,13 @@ import org.apache.tomcat.util.buf.CharChunk;
  * <p><strong>WARNING</strong> - There is a reported bug against the Netscape
  * provider code (com.netscape.jndi.ldap.LdapContextFactory) with respect to
  * successfully authenticated a non-existing user. The
- * report is here: http://issues.apache.org/bugzilla/show_bug.cgi?id=11210 .
+ * report is here: http://bz.apache.org/bugzilla/show_bug.cgi?id=11210 .
  * With luck, Netscape has updated their provider code and this is not an
  * issue. </p>
  *
  * @author John Holman
  * @author Craig R. McClanahan
- * @version $Revision: 769402 $ $Date: 2009-04-28 15:42:08 +0200 (Tue, 28 Apr 2009) $
+ *
  */
 
 public class JNDIRealm extends RealmBase {
@@ -197,14 +213,14 @@ public class JNDIRealm extends RealmBase {
      */
     protected String contextFactory = "com.sun.jndi.ldap.LdapCtxFactory";
 
-    
+
     /**
      * How aliases should be dereferenced during search operations.
      */
     protected String derefAliases = null;
 
     /**
-     * Constant that holds the name of the environment property for specifying 
+     * Constant that holds the name of the environment property for specifying
      * the manner in which aliases should be dereferenced.
      */
     public final static String DEREF_ALIASES = "java.naming.ldap.derefAliases";
@@ -230,9 +246,20 @@ public class JNDIRealm extends RealmBase {
 
 
     /**
-     * How should we handle referrals?  Microsoft Active Directory can't handle
-     * the default case, so an application authenticating against AD must
-     * set referrals to "follow".
+     * Should we ignore PartialResultExceptions when iterating over NamingEnumerations?
+     * Microsoft Active Directory often returns referrals, which lead
+     * to PartialResultExceptions. Unfortunately there's no stable way to detect,
+     * if the Exceptions really come from an AD referral.
+     * Set to true to ignore PartialResultExceptions.
+     */
+    protected boolean adCompat = false;
+
+
+    /**
+     * How should we handle referrals?  Microsoft Active Directory often returns
+     * referrals. If you need to follow them set referrals to "follow".
+     * Caution: if your DNS is not part of AD, the LDAP client lib might try
+     * to resolve your domain name in DNS to find another LDAP server.
      */
     protected String referrals = null;
 
@@ -294,7 +321,6 @@ public class JNDIRealm extends RealmBase {
      */
     protected MessageFormat[] userPatternFormatArray = null;
 
-
     /**
      * The base element for role searches.
      */
@@ -332,6 +358,12 @@ public class JNDIRealm extends RealmBase {
      * Should we search the entire subtree for matching memberships?
      */
     protected boolean roleSubtree = false;
+    
+    /**
+     * Should we look for nested group in order to determine roles?
+     */
+    protected boolean roleNested = false;
+
 
     /**
      * An alternate URL, to which, we should connect if connectionURL fails.
@@ -345,10 +377,17 @@ public class JNDIRealm extends RealmBase {
     protected int connectionAttempt = 0;
 
     /**
-     * The current user pattern to be used for lookup and binding of a user.
+     *  Add this role to every authenticated user
      */
-    protected int curUserPattern = 0;
+    protected String commonRole = null;
 
+
+    /**
+     * The timeout, in milliseconds, to use when trying to create a connection
+     * to the directory. The default is 5000 (5 seconds).
+     */
+    protected String connectionTimeout = "5000";
+    
     // ------------------------------------------------------------- Properties
 
     /**
@@ -463,11 +502,11 @@ public class JNDIRealm extends RealmBase {
      */
     public java.lang.String getDerefAliases() {
         return derefAliases;
-    }  
-    
+    }
+
     /**
      * Set the value for derefAliases to be used when searching the directory.
-     * 
+     *
      * @param derefAliases New value of property derefAliases.
      */
     public void setDerefAliases(java.lang.String derefAliases) {
@@ -495,6 +534,23 @@ public class JNDIRealm extends RealmBase {
     }
 
 
+    /**
+     * Returns the current settings for handling PartialResultExceptions
+     */
+    public boolean getAdCompat () {
+        return adCompat;
+    }
+
+
+    /**
+     * How do we handle PartialResultExceptions?
+     * True: ignore all PartialResultExceptions.
+     */
+    public void setAdCompat (boolean adCompat) {
+        this.adCompat = adCompat;
+    }
+
+
     /**
      * Returns the current settings for handling JNDI referrals.
      */
@@ -693,6 +749,28 @@ public class JNDIRealm extends RealmBase {
         this.roleSubtree = roleSubtree;
 
     }
+    
+    /**
+     * Return the "The nested group search flag" flag.
+     */
+    public boolean getRoleNested() {
+
+        return (this.roleNested);
+
+    }
+
+
+    /**
+     * Set the "search subtree for roles" flag.
+     *
+     * @param roleNested The nested group search flag
+     */
+    public void setRoleNested(boolean roleNested) {
+
+        this.roleNested = roleNested;
+
+    }
+     
 
 
     /**
@@ -778,6 +856,63 @@ public class JNDIRealm extends RealmBase {
     }
 
 
+    /**
+     * Return the common role
+     */
+    public String getCommonRole() {
+
+        return commonRole;
+
+    }
+
+
+    /**
+     * Set the common role
+     *
+     * @param commonRole The common role
+     */
+    public void setCommonRole(String commonRole) {
+
+        this.commonRole = commonRole;
+
+    }
+
+
+    /**
+     * Return the connection timeout.
+     */
+    public String getConnectionTimeout() {
+
+        return connectionTimeout;
+
+    }
+
+
+    /**
+     * Set the connection timeout.
+     *
+     * @param timeout The new connection timeout
+     */
+    public void setConnectionTimeout(String timeout) {
+
+        this.connectionTimeout = timeout;
+
+    }
+
+
+    /**
+     * Return descriptive information about this Realm implementation and
+     * the corresponding version number, in the format
+     * <code>&lt;description&gt;/&lt;version&gt;</code>.
+     */
+    @Override
+	public String getInfo() {
+
+        return info;
+
+    }
+
+
     // ---------------------------------------------------------- Realm Methods
 
 
@@ -877,6 +1012,8 @@ public class JNDIRealm extends RealmBase {
                 close(context);
 
             // Return "not authenticated" for this request
+            if (containerLog.isDebugEnabled())
+                containerLog.debug("Returning null principal.");
             return (null);
 
         }
@@ -907,21 +1044,30 @@ public class JNDIRealm extends RealmBase {
         throws NamingException {
 
         if (username == null || username.equals("")
-            || credentials == null || credentials.equals(""))
+            || credentials == null || credentials.equals("")) {
+            if (containerLog.isDebugEnabled())
+                containerLog.debug("username null or empty: returning null principal.");
             return (null);
+        }
 
         if (userPatternArray != null) {
-            for (curUserPattern = 0;
+            for (int curUserPattern = 0;
                  curUserPattern < userPatternFormatArray.length;
                  curUserPattern++) {
                 // Retrieve user information
-                User user = getUser(context, username);
+                User user = getUser(context, username, credentials, curUserPattern);
                 if (user != null) {
                     try {
                         // Check the user's credentials
                         if (checkCredentials(context, user, credentials)) {
                             // Search for additional roles
                             List<String> roles = getRoles(context, user);
+                            if (containerLog.isDebugEnabled()) {
+                                Iterator<String> it = roles.iterator();
+                                while (it.hasNext()) {
+                                    containerLog.debug("Found role: " + it.next());
+                                }
+                            }
                             return (new GenericPrincipal(this,
                                                          username,
                                                          credentials,
@@ -940,7 +1086,7 @@ public class JNDIRealm extends RealmBase {
             return null;
         } else {
             // Retrieve user information
-            User user = getUser(context, username);
+            User user = getUser(context, username, credentials);
             if (user == null)
                 return (null);
 
@@ -950,6 +1096,12 @@ public class JNDIRealm extends RealmBase {
 
             // Search for additional roles
             List<String> roles = getRoles(context, user);
+            if (containerLog.isDebugEnabled()) {
+                Iterator<String> it = roles.iterator();
+                while (it.hasNext()) {
+                    containerLog.debug("Found role: " + it.next());
+                }
+            }
 
             // Create and return a suitable Principal for this user
             return (new GenericPrincipal(this, username, credentials, roles));
@@ -957,6 +1109,45 @@ public class JNDIRealm extends RealmBase {
     }
 
 
+    /**
+     * Return a User object containing information about the user
+     * with the specified username, if found in the directory;
+     * otherwise return <code>null</code>.
+     *
+     * @param context The directory context
+     * @param username Username to be looked up
+     *
+     * @exception NamingException if a directory server error occurs
+     *
+     * @see #getUser(DirContext, String, String, int)
+     */
+    protected User getUser(DirContext context, String username)
+        throws NamingException {
+
+        return getUser(context, username, null, -1);
+    }
+
+
+    /**
+     * Return a User object containing information about the user
+     * with the specified username, if found in the directory;
+     * otherwise return <code>null</code>.
+     *
+     * @param context The directory context
+     * @param username Username to be looked up
+     * @param credentials User credentials (optional)
+     *
+     * @exception NamingException if a directory server error occurs
+     *
+     * @see #getUser(DirContext, String, String, int)
+     */
+    protected User getUser(DirContext context, String username, String credentials)
+        throws NamingException {
+
+        return getUser(context, username, credentials, -1);
+    }
+
+
     /**
      * Return a User object containing information about the user
      * with the specified username, if found in the directory;
@@ -970,10 +1161,13 @@ public class JNDIRealm extends RealmBase {
      *
      * @param context The directory context
      * @param username Username to be looked up
+     * @param credentials User credentials (optional)
+     * @param curUserPattern Index into userPatternFormatArray
      *
      * @exception NamingException if a directory server error occurs
      */
-    protected User getUser(DirContext context, String username)
+    protected User getUser(DirContext context, String username,
+                           String credentials, int curUserPattern)
         throws NamingException {
 
         User user = null;
@@ -988,8 +1182,8 @@ public class JNDIRealm extends RealmBase {
         list.toArray(attrIds);
 
         // Use pattern or search for user entry
-        if (userPatternFormatArray != null) {
-            user = getUserByPattern(context, username, attrIds);
+        if (userPatternFormatArray != null && curUserPattern >= 0) {
+            user = getUserByPattern(context, username, credentials, attrIds, curUserPattern);
         } else {
             user = getUserBySearch(context, username, attrIds);
         }
@@ -999,28 +1193,28 @@ public class JNDIRealm extends RealmBase {
 
 
     /**
-     * Use the <code>UserPattern</code> configuration attribute to
-     * locate the directory entry for the user with the specified
-     * username and return a User object; otherwise return
-     * <code>null</code>.
+     * Use the distinguished name to locate the directory
+     * entry for the user with the specified username and
+     * return a User object; otherwise return <code>null</code>.
      *
      * @param context The directory context
      * @param username The username
      * @param attrIds String[]containing names of attributes to
+     * @param dn Distinguished name of the user
      * retrieve.
      *
      * @exception NamingException if a directory server error occurs
      */
     protected User getUserByPattern(DirContext context,
-                                              String username,
-                                              String[] attrIds)
+                                    String username,
+                                    String[] attrIds,
+                                    String dn)
         throws NamingException {
 
-        if (username == null || userPatternFormatArray[curUserPattern] == null)
-            return (null);
-
-        // Form the dn from the user pattern
-        String dn = userPatternFormatArray[curUserPattern].format(new String[] { username });
+        // If no attributes are requested, no need to look for them
+        if (attrIds == null || attrIds.length == 0) {
+            return new User(username, dn, null, null);
+        }
 
         // Get required attributes from user entry
         Attributes attrs = null;
@@ -1046,6 +1240,71 @@ public class JNDIRealm extends RealmBase {
     }
 
 
+    /**
+     * Use the <code>UserPattern</code> configuration attribute to
+     * locate the directory entry for the user with the specified
+     * username and return a User object; otherwise return
+     * <code>null</code>.
+     *
+     * @param context The directory context
+     * @param username The username
+     * @param credentials User credentials (optional)
+     * @param attrIds String[]containing names of attributes to
+     * @param curUserPattern Index into userPatternFormatArray
+     *
+     * @exception NamingException if a directory server error occurs
+     * @see #getUserByPattern(DirContext, String, String[], String)
+     */
+    protected User getUserByPattern(DirContext context,
+                                    String username,
+                                    String credentials,
+                                    String[] attrIds,
+                                    int curUserPattern)
+        throws NamingException {
+
+        User user = null;
+
+        if (username == null || userPatternFormatArray[curUserPattern] == null)
+            return (null);
+
+        // Form the dn from the user pattern
+        String dn = userPatternFormatArray[curUserPattern].format(new String[] { username });
+
+        try {
+            user = getUserByPattern(context, username, attrIds, dn);
+        } catch (NameNotFoundException e) {
+            return (null);
+        } catch (NamingException e) {
+            // If the getUserByPattern() call fails, try it again with the
+            // credentials of the user that we're searching for
+            try {
+                // Set up security environment to bind as the user
+                context.addToEnvironment(Context.SECURITY_PRINCIPAL, dn);
+                context.addToEnvironment(Context.SECURITY_CREDENTIALS, credentials);
+
+                user = getUserByPattern(context, username, attrIds, dn);
+            } finally {
+                // Restore the original security environment
+                if (connectionName != null) {
+                    context.addToEnvironment(Context.SECURITY_PRINCIPAL,
+                                             connectionName);
+                } else {
+                    context.removeFromEnvironment(Context.SECURITY_PRINCIPAL);
+                }
+
+                if (connectionPassword != null) {
+                    context.addToEnvironment(Context.SECURITY_CREDENTIALS,
+                                             connectionPassword);
+                }
+                else {
+                    context.removeFromEnvironment(Context.SECURITY_CREDENTIALS);
+                }
+            }
+        }
+        return user;
+    }
+
+
     /**
      * Search the directory to return a User object containing
      * information about the user with the specified username, if
@@ -1058,8 +1317,8 @@ public class JNDIRealm extends RealmBase {
      * @exception NamingException if a directory server error occurs
      */
     protected User getUserBySearch(DirContext context,
-                                           String username,
-                                           String[] attrIds)
+                                   String username,
+                                   String[] attrIds)
         throws NamingException {
 
         if (username == null || userSearchFormat == null)
@@ -1083,56 +1342,63 @@ public class JNDIRealm extends RealmBase {
             attrIds = new String[0];
         constraints.setReturningAttributes(attrIds);
 
-        NamingEnumeration results =
+        NamingEnumeration<SearchResult> results =
             context.search(userBase, filter, constraints);
 
-
-        // Fail if no entries found
-        if (results == null || !results.hasMore()) {
-            return (null);
+        try {
+            // Fail if no entries found
+            try {
+                if (results == null || !results.hasMore()) {
+                    return null;
+                }
+            } catch (PartialResultException ex) {
+                if (!adCompat)
+                    throw ex;
+                else
+                    return null;
+            }
+    
+            // Get result for the first entry found
+            SearchResult result = results.next();
+    
+            // Check no further entries were found
+            try {
+                if (results.hasMore()) {
+                    if(containerLog.isInfoEnabled())
+                        containerLog.info("username " + username + " has multiple entries");
+                    return null;
+                }
+            } catch (PartialResultException ex) {
+                if (!adCompat)
+                    throw ex;
+            }
+    
+            String dn = getDistinguishedName(context, userBase, result);
+    
+            if (containerLog.isTraceEnabled())
+                containerLog.trace("  entry found for " + username + " with dn " + dn);
+    
+            // Get the entry's attributes
+            Attributes attrs = result.getAttributes();
+            if (attrs == null)
+                return (null);
+    
+            // Retrieve value of userPassword
+            String password = null;
+            if (userPassword != null)
+                password = getAttributeValue(userPassword, attrs);
+    
+            // Retrieve values of userRoleName attribute
+            ArrayList<String> roles = null;
+            if (userRoleName != null)
+                roles = addAttributeValues(userRoleName, attrs, roles);
+    
+            return new User(username, dn, password, roles);
+        } finally {
+            if (results != null) {
+                results.close();
+            }
         }
-
-        // Get result for the first entry found
-        SearchResult result = (SearchResult)results.next();
-
-        // Check no further entries were found
-        if (results.hasMore()) {
-            if(containerLog.isInfoEnabled())
-                containerLog.info("username " + username + " has multiple entries");
-            return (null);
-        }
-
-        // Get the entry's distinguished name
-        NameParser parser = context.getNameParser("");
-        Name contextName = parser.parse(context.getNameInNamespace());
-        Name baseName = parser.parse(userBase);
-
-        // Bugzilla 32269
-        Name entryName = parser.parse(new CompositeName(result.getName()).get(0));
-
-        Name name = contextName.addAll(baseName);
-        name = name.addAll(entryName);
-        String dn = name.toString();
-
-        if (containerLog.isTraceEnabled())
-            containerLog.trace("  entry found for " + username + " with dn " + dn);
-
-        // Get the entry's attributes
-        Attributes attrs = result.getAttributes();
-        if (attrs == null)
-            return null;
-
-        // Retrieve value of userPassword
-        String password = null;
-        if (userPassword != null)
-            password = getAttributeValue(userPassword, attrs);
-
-        // Retrieve values of userRoleName attribute
-        ArrayList<String> roles = null;
-        if (userRoleName != null)
-            roles = addAttributeValues(userRoleName, attrs, roles);
-
-        return new User(username, dn, password, roles);
     }
 
 
@@ -1167,10 +1433,10 @@ public class JNDIRealm extends RealmBase {
          if (containerLog.isTraceEnabled()) {
              if (validated) {
                  containerLog.trace(sm.getString("jndiRealm.authenticateSuccess",
-                                  user.username));
+                                  user.getUserName()));
              } else {
                  containerLog.trace(sm.getString("jndiRealm.authenticateFailure",
-                                  user.username));
+                                  user.getUserName()));
              }
          }
          return (validated);
@@ -1196,7 +1462,7 @@ public class JNDIRealm extends RealmBase {
         if (info == null || credentials == null)
             return (false);
 
-        String password = info.password;
+        String password = info.getPassword();
         if (password == null)
             return (false);
 
@@ -1206,10 +1472,10 @@ public class JNDIRealm extends RealmBase {
 
         boolean validated = false;
         if (hasMessageDigest()) {
-            // iPlanet support if the values starts with {SHA1}
+            // Some directories prefix the password with the hash type
             // The string is in a format compatible with Base64.encode not
             // the Hex encoding of the parent class.
-            if (password.startsWith("{SHA}")) {
+            if (password.startsWith("{MD5}") || password.startsWith("{SHA}")) {
                 /* sync since super.digest() does this same thing */
                 synchronized (this) {
                     password = password.substring(5);
@@ -1286,7 +1552,7 @@ public class JNDIRealm extends RealmBase {
          if (credentials == null || user == null)
              return (false);
 
-         String dn = user.dn;
+         String dn = user.getDN();
          if (dn == null)
              return (false);
 
@@ -1333,7 +1599,6 @@ public class JNDIRealm extends RealmBase {
         return (validated);
      }
 
-
     /**
      * Return a List of roles associated with the given User.  Any
      * roles present in the user's directory entry are supplemented by
@@ -1351,8 +1616,8 @@ public class JNDIRealm extends RealmBase {
         if (user == null)
             return (null);
 
-        String dn = user.dn;
-        String username = user.username;
+        String dn = user.getDN();
+        String username = user.getUserName();
 
         if (dn == null || username == null)
             return (null);
@@ -1361,15 +1626,24 @@ public class JNDIRealm extends RealmBase {
             containerLog.trace("  getRoles(" + dn + ")");
 
         // Start with roles retrieved from the user entry
-        ArrayList<String> list = user.roles;
-        if (list == null) {
-            list = new ArrayList<String>();
+        List<String> list = new ArrayList<String>(); 
+        List<String> userRoles = user.getRoles();
+        if (userRoles != null) {
+            list.addAll(userRoles); 
+        }
+        if (commonRole != null)
+            list.add(commonRole);
+
+        if (containerLog.isTraceEnabled()) {
+            containerLog.trace("  Found " + list.size() + " user internal roles");
+            for (int i=0; i<list.size(); i++)
+                containerLog.trace(  "  Found user internal role " + list.get(i));
         }
 
         // Are we configured to do role searches?
         if ((roleFormat == null) || (roleName == null))
             return (list);
-
+        
         // Set up parameters for an appropriate search
         String filter = roleFormat.format(new String[] { doRFC2254Encoding(dn), username });
         SearchControls controls = new SearchControls();
@@ -1380,30 +1654,91 @@ public class JNDIRealm extends RealmBase {
         controls.setReturningAttributes(new String[] {roleName});
 
         // Perform the configured search and process the results
-        NamingEnumeration results =
+        NamingEnumeration<SearchResult> results =
             context.search(roleBase, filter, controls);
         if (results == null)
             return (list);  // Should never happen, but just in case ...
-        while (results.hasMore()) {
-            SearchResult result = (SearchResult) results.next();
-            Attributes attrs = result.getAttributes();
-            if (attrs == null)
-                continue;
-            list = addAttributeValues(roleName, attrs, list);
+
+        HashMap<String, String> groupMap = new HashMap<String, String>();
+        try {
+            while (results.hasMore()) {
+                SearchResult result = results.next();
+                Attributes attrs = result.getAttributes();
+                if (attrs == null)
+                    continue;
+                String dname = getDistinguishedName(context, roleBase, result);
+                String name = getAttributeValue(roleName, attrs);
+                if (name != null && dname != null) {
+                    groupMap.put(dname, name);
+                }
+            }
+        } catch (PartialResultException ex) {
+            if (!adCompat)
+                throw ex;
+        } finally {
+            results.close();
         }
 
-
+        Set<String> keys = groupMap.keySet();
         if (containerLog.isTraceEnabled()) {
-            if (list != null) {
-                containerLog.trace("  Returning " + list.size() + " roles");
-                for (int i=0; i<list.size(); i++)
-                    containerLog.trace(  "  Found role " + list.get(i));
-            } else {
-                containerLog.trace("  getRoles about to return null ");
+            containerLog.trace("  Found " + keys.size() + " direct roles");
+            for (String key: keys) {
+                containerLog.trace(  "  Found direct role " + key + " -> " + groupMap.get(key));
             }
         }
 
-        return (list);
+        // if nested group search is enabled, perform searches for nested groups until no new group is found
+        if (getRoleNested()) {
+
+            // The following efficient algorithm is known as memberOf Algorithm, as described in "Practices in 
+            // Directory Groups". It avoids group slurping and handles cyclic group memberships as well.
+            // See http://middleware.internet2.edu/dir/ for details
+
+            Map<String, String> newGroups = new HashMap<String,String>(groupMap);
+            while (!newGroups.isEmpty()) {
+                Map<String, String> newThisRound = new HashMap<String, String>(); // Stores the groups we find in this iteration
+
+                for (Entry<String, String> group : newGroups.entrySet()) {
+                    filter = roleFormat.format(new String[] { group.getKey(), group.getValue() });
+
+                    if (containerLog.isTraceEnabled()) {
+                        containerLog.trace("Perform a nested group search with base "+ roleBase + " and filter " + filter);
+                    }
+
+                    results = context.search(roleBase, filter, controls);
+
+                    try {
+                        while (results.hasMore()) {
+                            SearchResult result = results.next();
+                            Attributes attrs = result.getAttributes();
+                            if (attrs == null)
+                                continue;
+                            String dname = getDistinguishedName(context, roleBase, result);
+                            String name = getAttributeValue(roleName, attrs);
+                            if (name != null && dname != null && !groupMap.keySet().contains(dname)) {
+                                groupMap.put(dname, name);
+                                newThisRound.put(dname, name);
+
+                                if (containerLog.isTraceEnabled()) {
+                                    containerLog.trace("  Found nested role " + dname + " -> " + name);
+                                }
+
+                            }
+                         }
+                    } catch (PartialResultException ex) {
+                        if (!adCompat)
+                            throw ex;
+                    } finally {
+                        results.close();
+                    }
+                }
+
+                newGroups = newThisRound;
+            }
+        }
+
+        list.addAll(groupMap.values());
+        return list;
     }
 
 
@@ -1464,10 +1799,17 @@ public class JNDIRealm extends RealmBase {
         Attribute attr = attrs.get(attrId);
         if (attr == null)
             return (values);
-        NamingEnumeration e = attr.getAll();
-        while(e.hasMore()) {
-            String value = (String)e.next();
-            values.add(value);
+        NamingEnumeration<?> e = attr.getAll();
+        try {
+            while(e.hasMore()) {
+                String value = (String)e.next();
+                values.add(value);
+            }
+        } catch (PartialResultException ex) {
+            if (!adCompat)
+                throw ex;
+        } finally {
+            e.close();
         }
         return values;
     }
@@ -1599,11 +1941,15 @@ public class JNDIRealm extends RealmBase {
     protected synchronized Principal getPrincipal(DirContext context,
                                                   String username)
         throws NamingException {
-        
+
         User user = getUser(context, username);
+
+        if (user != null) {
+            return new GenericPrincipal(this, user.getUserName(),
+                    user.getPassword(), getRoles(context, user));
+        }
         
-        return new GenericPrincipal(this, user.username, user.password ,
-                getRoles(context, user));
+        return null;
     }
 
     /**
@@ -1650,7 +1996,7 @@ public class JNDIRealm extends RealmBase {
      *
      * @return java.util.Hashtable the configuration for the directory context.
      */
-    protected Hashtable getDirectoryContextEnvironment() {
+    protected Hashtable<String,String> getDirectoryContextEnvironment() {
 
         Hashtable<String,String> env = new Hashtable<String,String>();
 
@@ -1676,6 +2022,8 @@ public class JNDIRealm extends RealmBase {
             env.put(Context.REFERRAL, referrals);
         if (derefAliases != null)
             env.put(JNDIRealm.DEREF_ALIASES, derefAliases);
+        if (connectionTimeout != null)
+        	env.put("com.sun.jndi.ldap.connect.timeout", connectionTimeout);
 
         return env;
 
@@ -1689,7 +2037,7 @@ public class JNDIRealm extends RealmBase {
      */
     protected void release(DirContext context) {
 
-        ; // NO-OP since we are not pooling anything
+        // NO-OP since we are not pooling anything
 
     }
 
@@ -1773,7 +2121,7 @@ public class JNDIRealm extends RealmBase {
                 startingPoint = endParenLoc+1;
                 startParenLoc = userPatternString.indexOf('(', startingPoint);
             }
-            return (String[])pathList.toArray(new String[] {});
+            return pathList.toArray(new String[] {});
         }
         return null;
 
@@ -1879,26 +2227,45 @@ public class JNDIRealm extends RealmBase {
     }
 
 
+     // ------------------------------------------------------ Private Classes
+    
+     /**
+      * A protected class representing a User
+      */
+     protected static class User {
+         
+         final private String username;
+         final private String dn;
+         final private String password;
+         final private List<String> roles;
+
+         public User(String username, String dn, String password,
+                 List<String> roles) {
+             this.username = username;
+             this.dn = dn;
+             this.password = password;
+             if (roles == null) {
+                 this.roles = Collections.emptyList();
+             } else {
+                 this.roles = Collections.unmodifiableList(roles);
+             }
+         }
+    
+         public String getUserName() {
+             return username;
+         }
+         
+         public String getDN() {
+             return dn;
+         }
+         
+         public String getPassword() {
+             return password;
+         }
+         
+         public List<String> getRoles() {
+             return roles;
+         }
+     }
 }
 
-// ------------------------------------------------------ Private Classes
-
-/**
- * A private class representing a User
- */
-class User {
-    String username = null;
-    String dn = null;
-    String password = null;
-    ArrayList<String> roles = null;
-
-
-    User(String username, String dn, String password,
-            ArrayList<String> roles) {
-        this.username = username;
-        this.dn = dn;
-        this.password = password;
-        this.roles = roles;
-    }
-
-}
