@@ -25,6 +25,7 @@ import java.io.DataInputStream;
 import java.io.File;
 import java.io.FileInputStream;
 import java.io.IOException;
+import java.io.UnsupportedEncodingException;
 import java.lang.reflect.Method;
 import java.security.AccessController;
 import java.security.MessageDigest;
@@ -40,6 +41,8 @@ import java.util.List;
 import java.util.Map;
 import java.util.Random;
 import java.util.concurrent.ConcurrentHashMap;
+import java.util.regex.Pattern;
+import java.util.regex.PatternSyntaxException;
 
 import javax.management.MBeanRegistration;
 import javax.management.MBeanServer;
@@ -66,16 +69,28 @@ import org.apache.tomcat.util.modeler.Registry;
  * be subclassed to create more sophisticated Manager implementations.
  *
  * @author Craig R. McClanahan
- * @version $Id: ManagerBase.java 1137629 2011-06-20 13:38:23Z markt $
+ *
  */
-
 public abstract class ManagerBase implements Manager, MBeanRegistration {
+
     protected Log log = LogFactory.getLog(ManagerBase.class);
 
     // ----------------------------------------------------- Instance Variables
 
+    private static final String devRandomSourceDefault;
+    static {
+        // - Use the default value only if it is a Unix-like system
+        // - Check that it exists 
+        File f = new File("/dev/urandom");
+        if (f.isAbsolute() && f.exists()) {
+            devRandomSourceDefault = f.getPath();
+        } else {
+            devRandomSourceDefault = null;
+        }
+    }
+
     protected DataInputStream randomIS=null;
-    protected String devRandomSource="/dev/urandom";
+    protected String devRandomSource = devRandomSourceDefault;
 
     /**
      * The default message digest algorithm to use if we cannot use
@@ -109,7 +124,11 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * The distributable flag for Sessions created by this Manager.  If this
      * flag is set to <code>true</code>, any user attributes added to a
      * session controlled by this Manager must be Serializable.
+     *
+     * @deprecated Ignored. {@link Context#getDistributable()} always takes
+     *             precedence. Will be removed in Tomcat 9.0.x.
      */
+    @Deprecated
     protected boolean distributable;
 
 
@@ -129,8 +148,12 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     /**
      * The default maximum inactive interval for Sessions created by
      * this Manager.
+     *
+     * @deprecated Ignored. {@link Context#getSessionTimeout()} always takes
+     *             precedence. Will be removed in Tomcat 9.0.x.
      */
-    protected int maxInactiveInterval = 60;
+    @Deprecated
+    protected int maxInactiveInterval = 30 * 60;
 
 
     /**
@@ -225,51 +248,165 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     /**
      * The string manager for this package.
      */
-    protected static StringManager sm =
-        StringManager.getManager(Constants.Package);
+    protected static final StringManager sm = StringManager.getManager(Constants.Package);
 
     /**
      * The property change support for this component.
      */
     protected PropertyChangeSupport support = new PropertyChangeSupport(this);
-    
+
+    private Pattern sessionAttributeNamePattern;
+
+    private Pattern sessionAttributeValueClassNamePattern;
+
+    private boolean warnOnSessionAttributeFilterFailure;
+
+
     // ------------------------------------------------------------- Security classes
 
-
     private class PrivilegedSetRandomFile
-            implements PrivilegedAction<DataInputStream>{
-        
+            implements PrivilegedAction<Void>{
+
+        private final String s;
+
         public PrivilegedSetRandomFile(String s) {
-            devRandomSource = s;
+            this.s = s;
         }
-        
-        public DataInputStream run(){
-            try {
-                File f=new File( devRandomSource );
-                if( ! f.exists() ) return null;
-                randomIS= new DataInputStream( new FileInputStream(f));
-                randomIS.readLong();
-                if( log.isDebugEnabled() )
-                    log.debug( "Opening " + devRandomSource );
-                return randomIS;
-            } catch (IOException ex){
-                log.warn("Error reading " + devRandomSource, ex);
-                if (randomIS != null) {
-                    try {
-                        randomIS.close();
-                    } catch (Exception e) {
-                        log.warn("Failed to close randomIS.");
-                    }
-                }
-                devRandomSource = null;
-                randomIS=null;
-                return null;
-            }
+
+        public Void run(){
+            doSetRandomFile(s);
+            return null;
         }
     }
 
 
-    // ------------------------------------------------------------- Properties
+    // ------------------------------------------------------------ Constructors
+
+    public ManagerBase() {
+        if (Globals.IS_SECURITY_ENABLED) {
+            // Minimum set required for default distribution/persistence to work
+            // plus String
+            setSessionAttributeValueClassNameFilter(
+                    "java\\.lang\\.(?:Boolean|Integer|Long|Number|String)");
+            setWarnOnSessionAttributeFilterFailure(true);
+        }
+    }
+
+
+    // -------------------------------------------------------------- Properties
+
+    /**
+     * Obtain the regular expression used to filter session attribute based on
+     * attribute name. The regular expression is anchored so it must match the
+     * entire name
+     *
+     * @return The regular expression currently used to filter attribute names.
+     *         {@code null} means no filter is applied. If an empty string is
+     *         specified then no names will match the filter and all attributes
+     *         will be blocked.
+     */
+    public String getSessionAttributeNameFilter() {
+        if (sessionAttributeNamePattern == null) {
+            return null;
+        }
+        return sessionAttributeNamePattern.toString();
+    }
+
+
+    public void setSessionAttributeNameFilter(String sessionAttributeNameFilter) {
+        if (sessionAttributeNameFilter == null || sessionAttributeNameFilter.length() == 0) {
+            sessionAttributeNamePattern = null;
+        } else {
+            sessionAttributeNamePattern = Pattern.compile(sessionAttributeNameFilter);
+        }
+    }
+
+
+    protected Pattern getSessionAttributeNamePattern() {
+        return sessionAttributeNamePattern;
+    }
+
+
+    /**
+     * Obtain the regular expression used to filter session attribute based on
+     * the implementation class of the value. The regular expression is anchored
+     * and must match the fully qualified class name.
+     *
+     * @return The regular expression currently used to filter class names.
+     *         {@code null} means no filter is applied. If an empty string is
+     *         specified then no names will match the filter and all attributes
+     *         will be blocked.
+     */
+    public String getSessionAttributeValueClassNameFilter() {
+        if (sessionAttributeValueClassNamePattern == null) {
+            return null;
+        }
+        return sessionAttributeValueClassNamePattern.toString();
+    }
+
+
+    /**
+     * Provides {@link #getSessionAttributeValueClassNameFilter()} as a
+     * pre-compiled regular expression pattern.
+     *
+     * @return The pre-compiled pattern used to filter session attributes based
+     *         on the implementation class name of the value. {@code null} means
+     *         no filter is applied.
+     */
+    protected Pattern getSessionAttributeValueClassNamePattern() {
+        return sessionAttributeValueClassNamePattern;
+    }
+
+
+    /**
+     * Set the regular expression to use to filter classes used for session
+     * attributes. The regular expression is anchored and must match the fully
+     * qualified class name.
+     *
+     * @param sessionAttributeValueClassNameFilter The regular expression to use
+     *            to filter session attributes based on class name. Use {@code
+     *            null} if no filtering is required. If an empty string is
+     *           specified then no names will match the filter and all
+     *           attributes will be blocked.
+     *
+     * @throws PatternSyntaxException If the expression is not valid
+     */
+    public void setSessionAttributeValueClassNameFilter(String sessionAttributeValueClassNameFilter)
+            throws PatternSyntaxException {
+        if (sessionAttributeValueClassNameFilter == null ||
+                sessionAttributeValueClassNameFilter.length() == 0) {
+            sessionAttributeValueClassNamePattern = null;
+        } else {
+            sessionAttributeValueClassNamePattern =
+                    Pattern.compile(sessionAttributeValueClassNameFilter);
+        }
+    }
+
+
+    /**
+     * Should a warn level log message be generated if a session attribute is
+     * not persisted / replicated / restored.
+     *
+     * @return {@code true} if a warn level log message should be generated
+     */
+    public boolean getWarnOnSessionAttributeFilterFailure() {
+        return warnOnSessionAttributeFilterFailure;
+    }
+
+
+    /**
+     * Configure whether or not a warn level log message should be generated if
+     * a session attribute is not persisted / replicated / restored.
+     *
+     * @param warnOnSessionAttributeFilterFailure {@code true} if the
+     *            warn level message should be generated
+     *
+     */
+    public void setWarnOnSessionAttributeFilterFailure(
+            boolean warnOnSessionAttributeFilterFailure) {
+        this.warnOnSessionAttributeFilterFailure = warnOnSessionAttributeFilterFailure;
+    }
+
 
     /**
      * Return the message digest algorithm for this Manager.
@@ -299,26 +436,27 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * Return the Container with which this Manager is associated.
      */
     public Container getContainer() {
-
-        return (this.container);
-
+        return this.container;
     }
 
 
-    /**
-     * Set the Container with which this Manager is associated.
-     *
-     * @param container The newly associated Container
-     */
     public void setContainer(Container container) {
-
+        if (this.container == container) {
+            // NO-OP
+            return;
+        }
+        if (initialized) {
+            throw new IllegalStateException(sm.getString("managerBase.setContextNotNew"));
+        }
         Container oldContainer = this.container;
         this.container = container;
+        // TODO - delete the line below in Tomcat 9 onwards
         support.firePropertyChange("container", oldContainer, this.container);
     }
 
 
-    /** Returns the name of the implementation class.
+    /**
+     * @return The name of the implementation class.
      */
     public String getClassName() {
         return this.getClass().getName();
@@ -364,28 +502,19 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * Return the distributable flag for the sessions supported by
      * this Manager.
      */
+    @Deprecated
     public boolean getDistributable() {
-
-        return (this.distributable);
-
+        Container container = getContainer();
+        if (container instanceof Context) {
+            return ((Context) container).getDistributable();
+        }
+        return false;
     }
 
 
-    /**
-     * Set the distributable flag for the sessions supported by this
-     * Manager.  If this flag is set, all user data objects added to
-     * sessions associated with this manager must implement Serializable.
-     *
-     * @param distributable The new distributable flag
-     */
+    @Deprecated
     public void setDistributable(boolean distributable) {
-
-        boolean oldDistributable = this.distributable;
-        this.distributable = distributable;
-        support.firePropertyChange("distributable",
-                                   new Boolean(oldDistributable),
-                                   new Boolean(this.distributable));
-
+        // NO-OP
     }
 
 
@@ -402,7 +531,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
             boolean apr = false;
             try {
                 String methodName = "random";
-                Class paramTypes[] = new Class[2];
+                Class<?> paramTypes[] = new Class[2];
                 paramTypes[0] = result.getClass();
                 paramTypes[1] = int.class;
                 Object paramValues[] = new Object[2];
@@ -416,7 +545,12 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
                 // Ignore
             }
             if (apr) {
-                setEntropy(new String(result));
+                try {
+                    setEntropy(new String(result, "ISO-8859-1"));
+                } catch (UnsupportedEncodingException ux) {
+                    // ISO-8859-1 should always be supported
+                    throw new Error(ux);
+                }
             } else {
                 setEntropy(this.toString());
             }
@@ -446,10 +580,9 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * the corresponding version number, in the format
      * <code>&lt;description&gt;/&lt;version&gt;</code>.
      */
+    @Deprecated
     public String getInfo() {
-
-        return (info);
-
+        return info;
     }
 
 
@@ -458,26 +591,18 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * for Sessions created by this Manager.
      */
     public int getMaxInactiveInterval() {
-
-        return (this.maxInactiveInterval);
-
+        Container container = getContainer();
+        if (container instanceof Context) {
+            // This method returns seconds, the Context uses minutes
+            return ((Context) container).getSessionTimeout() * 60;
+        }
+        return -1;
     }
 
 
-    /**
-     * Set the default maximum inactive interval (in seconds)
-     * for Sessions created by this Manager.
-     *
-     * @param interval The new default value
-     */
+    @Deprecated
     public void setMaxInactiveInterval(int interval) {
-
-        int oldMaxInactiveInterval = this.maxInactiveInterval;
-        this.maxInactiveInterval = interval;
-        support.firePropertyChange("maxInactiveInterval",
-                                   new Integer(oldMaxInactiveInterval),
-                                   new Integer(this.maxInactiveInterval));
-
+        log.warn(sm.getString("managerBase.setMaxInactiveIntervalUnused"));
     }
 
 
@@ -512,7 +637,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
 
 
     /**
-     * Return the descriptive short name of this Manager implementation.
+     * @return The descriptive short name of this Manager implementation.
      */
     public String getName() {
 
@@ -538,27 +663,49 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         // as a hack, you can use a static file - and generate the same
         // session ids ( good for strange debugging )
         if (Globals.IS_SECURITY_ENABLED){
-            randomIS = AccessController.doPrivileged(new PrivilegedSetRandomFile(s));
+            AccessController.doPrivileged(new PrivilegedSetRandomFile(s));
         } else {
-            try{
-                devRandomSource=s;
-                File f=new File( devRandomSource );
-                if( ! f.exists() ) return;
-                randomIS= new DataInputStream( new FileInputStream(f));
-                randomIS.readLong();
-                if( log.isDebugEnabled() )
-                    log.debug( "Opening " + devRandomSource );
-            } catch( IOException ex ) {
-                log.warn("Error reading " + devRandomSource, ex);
-                if (randomIS != null) {
-                    try {
-                        randomIS.close();
-                    } catch (Exception e) {
-                        log.warn("Failed to close randomIS.");
-                    }
+            doSetRandomFile(s);
+        }
+    }
+
+    private void doSetRandomFile(String s) {
+        DataInputStream is = null;
+        try {
+            if (s == null || s.length() == 0) {
+                return;
+            }
+            File f = new File(s);
+            if( ! f.exists() ) return;
+            if( log.isDebugEnabled() ) {
+                log.debug( "Opening " + s );
+            }
+            is = new DataInputStream( new FileInputStream(f));
+            is.readLong();
+        } catch( IOException ex ) {
+            log.warn("Error reading " + s, ex);
+            if (is != null) {
+                try {
+                    is.close();
+                } catch (Exception ex2) {
+                    log.warn("Failed to close " + s, ex2);
                 }
+                is = null;
+            }
+        } finally {
+            DataInputStream oldIS = randomIS;
+            if (is != null) {
+                devRandomSource = s;
+            } else {
                 devRandomSource = null;
-                randomIS=null;
+            }
+            randomIS = is;
+            if (oldIS != null) {
+                try {
+                    oldIS.close();
+                } catch (Exception ex) {
+                    log.warn("Failed to close RandomIS", ex);
+                }
             }
         }
     }
@@ -580,12 +727,12 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
             long t1 = seed;
             char entropy[] = getEntropy().toCharArray();
             for (int i = 0; i < entropy.length; i++) {
-                long update = ((byte) entropy[i]) << ((i % 8) * 8);
+                long update = ((long) entropy[i]) << ((i % 8) * 8);
                 seed ^= update;
             }
             try {
                 // Construct and seed a new random number generator
-                Class clazz = Class.forName(randomClass);
+                Class<?> clazz = Class.forName(randomClass);
                 this.random = (Random) clazz.newInstance();
                 this.random.setSeed(seed);
             } catch (Exception e) {
@@ -602,8 +749,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
             }
         }
         
-        return (this.random);
-
+        return this.random;
     }
 
 
@@ -611,9 +757,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * Return the random number generator class name.
      */
     public String getRandomClass() {
-
-        return (this.randomClass);
-
+        return this.randomClass;
     }
 
 
@@ -623,12 +767,9 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * @param randomClass The new random number generator class name
      */
     public void setRandomClass(String randomClass) {
-
         String oldRandomClass = this.randomClass;
         this.randomClass = randomClass;
-        support.firePropertyChange("randomClass", oldRandomClass,
-                                   this.randomClass);
-
+        support.firePropertyChange("randomClass", oldRandomClass, this.randomClass);
     }
 
 
@@ -661,12 +802,10 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     }
     
     /**
-     * Return the frequency of manager checks.
+     * @return The frequency of manager checks.
      */
     public int getProcessExpiresFrequency() {
-
-        return (this.processExpiresFrequency);
-
+        return this.processExpiresFrequency;
     }
 
     /**
@@ -685,11 +824,11 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         support.firePropertyChange("processExpiresFrequency",
                                    new Integer(oldProcessExpiresFrequency),
                                    new Integer(this.processExpiresFrequency));
-
     }
+    
+    
     // --------------------------------------------------------- Public Methods
 
-
     /**
      * Implements the Manager interface, direct call to processExpires
      */
@@ -747,7 +886,6 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         if( oname==null ) {
             try {
                 StandardContext ctx=(StandardContext)this.getContainer();
-                Engine eng=(Engine)ctx.getParent().getParent();
                 domain=ctx.getEngineName();
                 distributable = ctx.getDistributable();
                 StandardHost hst=(StandardHost)ctx.getParent();
@@ -766,6 +904,10 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         // Initialize random number generation
         getRandomBytes(new byte[16]);
         
+        if (!(container instanceof Context)) {
+            throw new IllegalStateException(sm.getString("managerBase.contextNull"));
+        }
+
         // Ensure caches for timing stats are the right size by filling with
         // nulls.
         while (sessionCreationTiming.size() < TIMING_STATS_CACHE_SIZE) {
@@ -822,6 +964,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      *  instantiated for any reason
      * @deprecated
      */
+    @Deprecated
     public Session createSession() {
         return createSession(null);
     }
@@ -849,7 +992,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         session.setNew(true);
         session.setValid(true);
         session.setCreationTime(System.currentTimeMillis());
-        session.setMaxInactiveInterval(this.maxInactiveInterval);
+        session.setMaxInactiveInterval(((Context) getContainer()).getSessionTimeout() * 60);
         if (sessionId == null) {
             sessionId = generateSessionId();
         // FIXME WHy we need no duplication check?
@@ -915,11 +1058,10 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      *  processing this request
      */
     public Session findSession(String id) throws IOException {
-
-        if (id == null)
-            return (null);
-        return (Session) sessions.get(id);
-
+        if (id == null) {
+            return null;
+        }
+        return sessions.get(id);
     }
 
 
@@ -928,9 +1070,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * If this Manager has no active Sessions, a zero-length array is returned.
      */
     public Session[] findSessions() {
-
         return sessions.values().toArray(new Session[0]);
-
     }
 
 
@@ -940,9 +1080,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * @param session Session to be removed
      */
     public void remove(Session session) {
-
         sessions.remove(session.getIdInternal());
-
     }
 
 
@@ -952,9 +1090,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * @param listener The listener to remove
      */
     public void removePropertyChangeListener(PropertyChangeListener listener) {
-
         support.removePropertyChangeListener(listener);
-
     }
 
 
@@ -976,8 +1112,53 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     }
     
     
-    // ------------------------------------------------------ Protected Methods
+    /**
+     * {@inheritDoc}
+     * <p>
+     * This implementation excludes session attributes from distribution if the:
+     * <ul>
+     * <li>attribute name matches {@link #getSessionAttributeNameFilter()}</li>
+     * </ul>
+     */
+    public boolean willAttributeDistribute(String name, Object value) {
+        Pattern sessionAttributeNamePattern = getSessionAttributeNamePattern();
+        if (sessionAttributeNamePattern != null) {
+            if (!sessionAttributeNamePattern.matcher(name).matches()) {
+                if (getWarnOnSessionAttributeFilterFailure() || log.isDebugEnabled()) {
+                    String msg = sm.getString("managerBase.sessionAttributeNameFilter",
+                            name, sessionAttributeNamePattern);
+                    if (getWarnOnSessionAttributeFilterFailure()) {
+                        log.warn(msg);
+                    } else {
+                        log.debug(msg);
+                    }
+                }
+                return false;
+            }
+        }
 
+        Pattern sessionAttributeValueClassNamePattern = getSessionAttributeValueClassNamePattern();
+        if (value != null && sessionAttributeValueClassNamePattern != null) {
+            if (!sessionAttributeValueClassNamePattern.matcher(
+                    value.getClass().getName()).matches()) {
+                if (getWarnOnSessionAttributeFilterFailure() || log.isDebugEnabled()) {
+                    String msg = sm.getString("managerBase.sessionAttributeValueClassNameFilter",
+                            name, value.getClass().getName(), sessionAttributeNamePattern);
+                    if (getWarnOnSessionAttributeFilterFailure()) {
+                        log.warn(msg);
+                    } else {
+                        log.debug(msg);
+                    }
+                }
+                return false;
+            }
+        }
+
+        return true;
+    }
+
+
+    // ------------------------------------------------------ Protected Methods
 
     /**
      * Get new session class to be used in the doLoad() method.
@@ -1075,7 +1256,7 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     public Engine getEngine() {
         Engine e = null;
         for (Container c = getContainer(); e == null && c != null ; c = c.getParent()) {
-            if (c != null && c instanceof Engine) {
+            if (c instanceof Engine) {
                 e = (Engine)c;
             }
         }
@@ -1288,12 +1469,13 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
 
 
     /** 
-     * For debugging: return a list of all session ids currently active
+     * For debugging.
      *
+     * @return A space separated list of all session IDs currently active
      */
     public String listSessionIds() {
-        StringBuffer sb=new StringBuffer();
-        Iterator keys = sessions.keySet().iterator();
+        StringBuilder sb = new StringBuilder();
+        Iterator<String> keys = sessions.keySet().iterator();
         while (keys.hasNext()) {
             sb.append(keys.next()).append(" ");
         }
@@ -1302,14 +1484,16 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
 
 
     /** 
-     * For debugging: get a session attribute
+     * For debugging.
      *
-     * @param sessionId
-     * @param key
-     * @return The attribute value, if found, null otherwise
+     * @param sessionId The ID for the session of interest
+     * @param key       The key for the attribute to obtain
+     *
+     * @return The attribute value for the specified session, if found, null
+     *         otherwise
      */
     public String getSessionAttribute( String sessionId, String key ) {
-        Session s = (Session) sessions.get(sessionId);
+        Session s = sessions.get(sessionId);
         if( s==null ) {
             if(log.isInfoEnabled())
                 log.info("Session not found " + sessionId);
@@ -1333,8 +1517,8 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
      * representation of their values, or null if no session with the
      * specified id exists, or if the session does not have any attributes
      */
-    public HashMap getSession(String sessionId) {
-        Session s = (Session) sessions.get(sessionId);
+    public HashMap<String,Object> getSession(String sessionId) {
+        Session s = sessions.get(sessionId);
         if (s == null) {
             if (log.isInfoEnabled()) {
                 log.info("Session not found " + sessionId);
@@ -1342,14 +1526,15 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
             return null;
         }
 
-        Enumeration ee = s.getSession().getAttributeNames();
+        @SuppressWarnings("unchecked")
+        Enumeration<String> ee = s.getSession().getAttributeNames();
         if (ee == null || !ee.hasMoreElements()) {
             return null;
         }
 
-        HashMap map = new HashMap();
+        HashMap<String,Object> map = new HashMap<String,Object>();
         while (ee.hasMoreElements()) {
-            String attrName = (String) ee.nextElement();
+            String attrName = ee.nextElement();
             map.put(attrName, getSessionAttribute(sessionId, attrName));
         }
 
@@ -1358,8 +1543,8 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
 
 
     public void expireSession( String sessionId ) {
-        Session s=(Session)sessions.get(sessionId);
-        if( s==null ) {
+        Session s = sessions.get(sessionId);
+        if (s == null) {
             if(log.isInfoEnabled())
                 log.info("Session not found " + sessionId);
             return;
@@ -1368,15 +1553,16 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     }
 
     public long getLastAccessedTimestamp( String sessionId ) {
-        Session s=(Session)sessions.get(sessionId);
-        if(s== null)
-            return -1 ;
+        Session s = sessions.get(sessionId);
+        if (s == null) {
+            return -1;
+        }
         return s.getLastAccessedTime();
     }
   
     public String getLastAccessedTime( String sessionId ) {
-        Session s=(Session)sessions.get(sessionId);
-        if( s==null ) {
+        Session s = sessions.get(sessionId);
+        if (s == null) {
             if(log.isInfoEnabled())
                 log.info("Session not found " + sessionId);
             return "";
@@ -1385,8 +1571,8 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     }
 
     public String getCreationTime( String sessionId ) {
-        Session s=(Session)sessions.get(sessionId);
-        if( s==null ) {
+        Session s = sessions.get(sessionId);
+        if (s == null) {
             if(log.isInfoEnabled())
                 log.info("Session not found " + sessionId);
             return "";
@@ -1395,9 +1581,10 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
     }
 
     public long getCreationTimestamp( String sessionId ) {
-        Session s=(Session)sessions.get(sessionId);
-        if(s== null)
-            return -1 ;
+        Session s = sessions.get(sessionId);
+        if (s == null) {
+            return -1;
+        }
         return s.getCreationTime();
     }
 
@@ -1443,15 +1630,16 @@ public abstract class ManagerBase implements Manager, MBeanRegistration {
         }
         
         /**
-         * Time stamp associated with this piece of timing information in
-         * milliseconds.
+         * @return Time stamp associated with this piece of timing information
+         *         in milliseconds.
          */
         public long getTimestamp() {
             return timestamp;
         }
         
         /**
-         * Duration associated with this piece of timing information in seconds.
+         * @return Duration associated with this piece of timing information in
+         *         seconds.
          */
         public int getDuration() {
             return duration;
