@@ -35,6 +35,7 @@ import java.util.Hashtable;
 import java.util.Iterator;
 import java.util.Map;
 import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.atomic.AtomicInteger;
 
 import javax.servlet.ServletContext;
@@ -60,6 +61,7 @@ import org.apache.catalina.core.StandardContext;
 import org.apache.catalina.realm.GenericPrincipal;
 import org.apache.catalina.security.SecurityUtil;
 import org.apache.catalina.session.ManagerBase.SessionTiming;
+import org.apache.tomcat.util.security.PrivilegedSetTccl;
 
 /**
  * Standard implementation of the <b>Session</b> interface.  This object is
@@ -79,7 +81,7 @@ import org.apache.catalina.session.ManagerBase.SessionTiming;
  * @author Craig R. McClanahan
  * @author Sean Legassick
  * @author <a href="mailto:jon@latchkey.com">Jon S. Stevens</a>
- * @version $Id: StandardSession.java 1135497 2011-06-14 11:41:45Z markt $
+ *
  */
 
 public class StandardSession
@@ -132,7 +134,7 @@ public class StandardSession
     /**
      * The collection of user data attributes associated with this Session.
      */
-    protected Map attributes = new ConcurrentHashMap();
+    protected ConcurrentMap<String, Object> attributes = new ConcurrentHashMap<String, Object>();
 
 
     /**
@@ -152,7 +154,11 @@ public class StandardSession
 
     /**
      * Set of attribute names which are not allowed to be persisted.
+     *
+     * @deprecated Use {@link Constants#excludedAttributeNames} instead. Will be
+     *             removed in Tomcat 9.
      */
+    @Deprecated
     protected static final String[] excludedAttributes = {
         Globals.SUBJECT_ATTR
     };
@@ -242,8 +248,7 @@ public class StandardSession
     /**
      * The string manager for this package.
      */
-    protected static StringManager sm =
-        StringManager.getManager(Constants.Package);
+    protected static final StringManager sm = StringManager.getManager(Constants.Package);
 
 
     /**
@@ -450,9 +455,7 @@ public class StandardSession
      * Return the Manager within which this Session is valid.
      */
     public Manager getManager() {
-
-        return (this.manager);
-
+        return this.manager;
     }
 
 
@@ -462,9 +465,7 @@ public class StandardSession
      * @param manager The new Manager
      */
     public void setManager(Manager manager) {
-
         this.manager = manager;
-
     }
 
 
@@ -1042,15 +1043,11 @@ public class StandardSession
      * Return the ServletContext to which this session belongs.
      */
     public ServletContext getServletContext() {
-
-        if (manager == null)
-            return (null);
+        if (manager == null) {
+            return null;
+        }
         Context context = (Context) manager.getContainer();
-        if (context == null)
-            return (null);
-        else
-            return (context.getServletContext());
-
+        return context.getServletContext();
     }
 
 
@@ -1343,13 +1340,15 @@ public class StandardSession
         }
 
         // Validate our current state
-        if (!isValidInternal())
+        if (!isValidInternal()) {
             throw new IllegalStateException
                 (sm.getString("standardSession.setAttribute.ise"));
-        if ((manager != null) && manager.getDistributable() &&
-          !(value instanceof Serializable))
-            throw new IllegalArgumentException
-                (sm.getString("standardSession.setAttribute.iae", name));
+        }
+        if ((manager != null) && ((Context) manager.getContainer()).getDistributable() &&
+                !isAttributeDistributable(name, value) && !exclude(name, value)) {
+            throw new IllegalArgumentException(sm.getString(
+                    "standardSession.setAttribute.iae", name));
+        }
         // Construct an event with the new value
         HttpSessionBindingEvent event = null;
 
@@ -1454,6 +1453,17 @@ public class StandardSession
         return (this.isValid || this.expiring);
     }
 
+    /**
+     * {@inheritDoc}
+     * <p>
+     * This implementation simply checks the value for serializability.
+     * Sub-classes might use other distribution technology not based on
+     * serialization and can override this check.
+     */
+    public boolean isAttributeDistributable(String name, Object value) {
+        return value instanceof Serializable;
+    }
+
 
     /**
      * Read a serialized version of this session object from the specified
@@ -1487,7 +1497,7 @@ public class StandardSession
 
         // Deserialize the attribute count and attribute values
         if (attributes == null)
-            attributes = new Hashtable();
+            attributes = new ConcurrentHashMap<String,Object>();
         int n = ((Integer) stream.readObject()).intValue();
         boolean isValidSave = isValid;
         isValid = true;
@@ -1499,6 +1509,11 @@ public class StandardSession
             if (manager.getContainer().getLogger().isDebugEnabled())
                 manager.getContainer().getLogger().debug("  loading attribute '" + name +
                     "' with value '" + value + "'");
+            // Handle the case where the filter configuration was changed while
+            // the web application was stopped.
+            if (exclude(name, value)) {
+                continue;
+            }
             attributes.put(name, value);
         }
         isValid = isValidSave;
@@ -1552,10 +1567,9 @@ public class StandardSession
         ArrayList saveValues = new ArrayList();
         for (int i = 0; i < keys.length; i++) {
             Object value = attributes.get(keys[i]);
-            if (value == null)
+            if (value == null) {
                 continue;
-            else if ( (value instanceof Serializable) 
-                    && (!exclude(keys[i]) )) {
+            } else if (isAttributeDistributable(keys[i], value) && !exclude(keys[i], value)) {
                 saveNames.add(keys[i]);
                 saveValues.add(value);
             } else {
@@ -1590,17 +1604,51 @@ public class StandardSession
 
 
     /**
-     * Exclude attribute that cannot be serialized.
+     * Exclude standard attributes that cannot be serialized.
      * @param name the attribute's name
+     *
+     * @deprecated Use {@link #exclude(String, Object)}. Will be removed in
+     *             Tomcat 9.0.x.
      */
+    @Deprecated
     protected boolean exclude(String name){
+        return exclude(name, null);
+    }
 
-        for (int i = 0; i < excludedAttributes.length; i++) {
-            if (name.equalsIgnoreCase(excludedAttributes[i]))
-                return true;
+
+    /**
+     * Should the given session attribute be excluded? This implementation
+     * checks:
+     * <ul>
+     * <li>{@link Constants#excludedAttributeNames}</li>
+     * <li>{@link Manager#willAttributeDistribute(String, Object)}</li>
+     * </ul>
+     * Note: This method deliberately does not check
+     *       {@link #isAttributeDistributable(String, Object)} which is kept
+     *       separate to support the checks required in
+     *       {@link #setAttribute(String, Object, boolean)}
+     *
+     * @param name  The attribute name
+     * @param value The attribute value
+     *
+     * @return {@code true} if the attribute should be excluded from
+     *         distribution, otherwise {@code false}
+     */
+    protected boolean exclude(String name, Object value) {
+        if (Constants.excludedAttributeNames.contains(name)) {
+            return true;
         }
 
-        return false;
+        // Manager is required for remaining check
+        Manager manager = getManager();
+        if (manager == null) {
+            // Manager may be null during replication of new sessions in a
+            // cluster. Avoid the NPE.
+            return false;
+        }
+
+        // Last check so use a short-cut
+        return !manager.willAttributeDistribute(name, value);
     }
 
 
@@ -1731,28 +1779,9 @@ public class StandardSession
                     (sm.getString("standardSession.attributeEvent"), t);
             }
         }
-
     }
-
-
-    private static class PrivilegedSetTccl
-    implements PrivilegedAction<Void> {
-
-        private ClassLoader cl;
-
-        PrivilegedSetTccl(ClassLoader cl) {
-            this.cl = cl;
-        }
-
-        public Void run() {
-            Thread.currentThread().setContextClassLoader(cl);
-            return null;
-        }
-    }
-
 }
 
-
 // ------------------------------------------------------------ Protected Class
 
 
