@@ -18,9 +18,9 @@
 package org.apache.coyote.http11;
 
 import java.net.InetAddress;
-import java.net.URLEncoder;
 import java.util.HashMap;
 import java.util.Iterator;
+import java.util.Map;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.Executor;
@@ -31,13 +31,13 @@ import javax.management.MBeanRegistration;
 import javax.management.MBeanServer;
 import javax.management.ObjectName;
 
+import org.apache.coyote.AbstractProtocol;
 import org.apache.coyote.ActionCode;
-import org.apache.coyote.ActionHook;
 import org.apache.coyote.Adapter;
-import org.apache.coyote.ProtocolHandler;
 import org.apache.coyote.RequestGroupInfo;
 import org.apache.coyote.RequestInfo;
 import org.apache.tomcat.util.modeler.Registry;
+import org.apache.tomcat.util.net.AbstractEndpoint;
 import org.apache.tomcat.util.net.AprEndpoint;
 import org.apache.tomcat.util.net.SocketStatus;
 import org.apache.tomcat.util.net.AprEndpoint.Handler;
@@ -52,9 +52,10 @@ import org.apache.tomcat.util.res.StringManager;
  * @author Remy Maucherat
  * @author Costin Manolache
  */
-public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
+public class Http11AprProtocol extends AbstractProtocol
+    implements MBeanRegistration {
 
-    protected static org.apache.juli.logging.Log log = 
+    protected static org.apache.juli.logging.Log log =
         org.apache.juli.logging.LogFactory.getLog(Http11AprProtocol.class);
 
     /**
@@ -85,7 +86,7 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
         return attributes.get(key);
     }
 
-    public Iterator getAttributeNames() {
+    public Iterator<String> getAttributeNames() {
         return attributes.keySet().iterator();
     }
 
@@ -176,18 +177,16 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
     }
 
     public String getName() {
-        String encodedAddr = "";
-        if (getAddress() != null) {
-            encodedAddr = "" + getAddress();
-            if (encodedAddr.startsWith("/"))
-                encodedAddr = encodedAddr.substring(1);
-            encodedAddr = URLEncoder.encode(encodedAddr) + "-";
-        }
-        return ("http-" + encodedAddr + endpoint.getPort());
+        return createName("http", getAddress(), endpoint.getPort());
     }
 
     protected AprEndpoint endpoint=new AprEndpoint();
 
+    @Override
+    protected final AbstractEndpoint getEndpoint() {
+        return endpoint;
+    }
+
     protected HashMap<String, Object> attributes = new HashMap<String, Object>();
 
     private Http11ConnectionHandler cHandler = new Http11ConnectionHandler(this);
@@ -201,7 +200,7 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
 
     public Executor getExecutor() { return endpoint.getExecutor(); }
     public void setExecutor(Executor executor) { endpoint.setExecutor(executor); }
-    
+
     public int getMaxThreads() { return endpoint.getMaxThreads(); }
     public void setMaxThreads(int maxThreads) { endpoint.setMaxThreads(maxThreads); }
 
@@ -226,6 +225,11 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
     public int getSoTimeout() { return endpoint.getSoTimeout(); }
     public void setSoTimeout(int soTimeout) { endpoint.setSoTimeout(soTimeout); }
 
+    public int getUnlockTimeout() { return endpoint.getUnlockTimeout(); }
+    public void setUnlockTimeout(int unlockTimeout) {
+        endpoint.setUnlockTimeout(unlockTimeout);
+    }
+
     /**
      * The number of seconds Tomcat will wait for a subsequent request
      * before closing the connection.
@@ -244,10 +248,10 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
 
     public void setPollerThreadCount(int pollerThreadCount) { endpoint.setPollerThreadCount(pollerThreadCount); }
     public int getPollerThreadCount() { return endpoint.getPollerThreadCount(); }
-    
+
     public int getSendfileSize() { return endpoint.getSendfileSize(); }
     public void setSendfileSize(int sendfileSize) { endpoint.setSendfileSize(sendfileSize); }
-    
+
     public void setSendfileThreadCount(int sendfileThreadCount) { endpoint.setSendfileThreadCount(sendfileThreadCount); }
     public int getSendfileThreadCount() { return endpoint.getSendfileThreadCount(); }
 
@@ -291,20 +295,20 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
     protected String compression = "off";
     public String getCompression() { return compression; }
     public void setCompression(String valueS) { compression = valueS; }
-    
-    
+
+
     // HTTP
     protected String noCompressionUserAgents = null;
     public String getNoCompressionUserAgents() { return noCompressionUserAgents; }
     public void setNoCompressionUserAgents(String valueS) { noCompressionUserAgents = valueS; }
 
-    
+
     // HTTP
     protected String compressableMimeTypes = "text/html,text/xml,text/plain";
     public String getCompressableMimeType() { return compressableMimeTypes; }
     public void setCompressableMimeType(String valueS) { compressableMimeTypes = valueS; }
-    
-    
+
+
     // HTTP
     protected int compressionMinSize = 2048;
     public int getCompressionMinSize() { return compressionMinSize; }
@@ -318,14 +322,14 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
     protected String restrictedUserAgents = null;
     public String getRestrictedUserAgents() { return restrictedUserAgents; }
     public void setRestrictedUserAgents(String valueS) { restrictedUserAgents = valueS; }
-    
-    
+
+
     protected String protocol = null;
     public String getProtocol() { return protocol; }
     public void setProtocol(String protocol) { setSecure(true); this.protocol = protocol; }
 
     /**
-     * Maximum number of requests which can be performed over a keepalive 
+     * Maximum number of requests which can be performed over a keepalive
      * connection. The default is the same as for Apache HTTP Server.
      */
     protected int maxKeepAliveRequests = 100;
@@ -404,6 +408,17 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
     public void setSSLCipherSuite(String SSLCipherSuite) { endpoint.setSSLCipherSuite(SSLCipherSuite); }
 
 
+    /**
+     * SSL honor cipher order.
+	 *
+     * Set to <code>true</code> to enforce the <i>server's</i> cipher order
+     * instead of the default which is to allow the client to choose a
+     * preferred cipher.
+     */
+    public boolean getSSLHonorCipherOrder() { return endpoint.getSSLHonorCipherOrder(); }
+    public void setSSLHonorCipherOrder(boolean SSLHonorCipherOrder) { endpoint.setSSLHonorCipherOrder(SSLHonorCipherOrder); }
+
+
     /**
      * SSL certificate file.
      */
@@ -465,20 +480,43 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
      */
     public int getSSLVerifyDepth() { return endpoint.getSSLVerifyDepth(); }
     public void setSSLVerifyDepth(int SSLVerifyDepth) { endpoint.setSSLVerifyDepth(SSLVerifyDepth); }
-    
+
+    /**
+     * Disable SSL compression.
+     */
+    public boolean getSSLDisableCompression() { return endpoint.getSSLDisableCompression(); }
+    public void setSSLDisableCompression(boolean disable) { endpoint.setSSLDisableCompression(disable); }
+
+
+    /**
+     * When client certificate information is presented in a form other than
+     * instances of {@link java.security.cert.X509Certificate} it needs to be
+     * converted before it can be used and this property controls which JSSE
+     * provider is used to perform the conversion. For example it is used with
+     * the AJP connectors, the HTTP APR connector and with the
+     * {@link org.apache.catalina.valves.SSLValve}. If not specified, the
+     * default provider will be used.
+     */
+    protected String clientCertProvider = null;
+    public String getClientCertProvider() { return clientCertProvider; }
+    public void setClientCertProvider(String s) { this.clientCertProvider = s; }
+
+
     // --------------------  Connection handler --------------------
 
     static class Http11ConnectionHandler implements Handler {
-        
+
         protected Http11AprProtocol proto;
         protected AtomicLong registerCount = new AtomicLong(0);
         protected RequestGroupInfo global = new RequestGroupInfo();
-        
-        protected ConcurrentHashMap<Long, Http11AprProcessor> connections =
-            new ConcurrentHashMap<Long, Http11AprProcessor>();
-        protected ConcurrentLinkedQueue<Http11AprProcessor> recycledProcessors = 
+
+        protected final Map<Long, Http11AprProcessor> connections =
+                new ConcurrentHashMap<Long, Http11AprProcessor>();
+        protected ConcurrentLinkedQueue<Http11AprProcessor> recycledProcessors =
             new ConcurrentLinkedQueue<Http11AprProcessor>() {
+            private static final long serialVersionUID = 1L;
             protected AtomicInteger size = new AtomicInteger(0);
+            @Override
             public boolean offer(Http11AprProcessor processor) {
                 boolean offer = (proto.processorCache == -1) ? true : (size.get() < proto.processorCache);
                 //avoid over growing our cache or add after we have stopped
@@ -492,7 +530,8 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
                 if (!result) unregister(processor);
                 return result;
             }
-            
+
+            @Override
             public Http11AprProcessor poll() {
                 Http11AprProcessor result = super.poll();
                 if ( result != null ) {
@@ -500,7 +539,8 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
                 }
                 return result;
             }
-            
+
+            @Override
             public void clear() {
                 Http11AprProcessor next = poll();
                 while ( next != null ) {
@@ -518,9 +558,9 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
         }
 
         public SocketState event(long socket, SocketStatus status) {
-            Http11AprProcessor result = connections.get(socket);
-            
-            SocketState state = SocketState.CLOSED; 
+            Http11AprProcessor result = connections.get(Long.valueOf(socket));
+
+            SocketState state = SocketState.CLOSED;
             if (result != null) {
                 // Call the appropriate event
                 try {
@@ -547,7 +587,7 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
                         (sm.getString("http11protocol.proto.error"), e);
                 } finally {
                     if (state != SocketState.LONG) {
-                        connections.remove(socket);
+                        connections.remove(Long.valueOf(socket));
                         recycledProcessors.offer(result);
                         if (state == SocketState.OPEN) {
                             proto.endpoint.getPoller().add(socket);
@@ -559,25 +599,40 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
             }
             return state;
         }
-        
+
         public SocketState process(long socket) {
-            Http11AprProcessor processor = recycledProcessors.poll();
+            Http11AprProcessor processor = null;
             try {
+                processor = connections.remove(Long.valueOf(socket));
+
+                if (processor == null) {
+                    processor = recycledProcessors.poll();
+                }
+
                 if (processor == null) {
                     processor = createProcessor();
                 }
 
-                if (processor instanceof ActionHook) {
-                    ((ActionHook) processor).action(ActionCode.ACTION_START, null);
-                }
+                processor.action(ActionCode.ACTION_START, null);
 
                 SocketState state = processor.process(socket);
                 if (state == SocketState.LONG) {
-                    // Associate the connection with the processor. The next request 
+                    // Associate the connection with the processor. The next request
                     // processed by this thread will use either a new or a recycled
                     // processor.
-                    connections.put(socket, processor);
+                    connections.put(Long.valueOf(socket), processor);
                     proto.endpoint.getCometPoller().add(socket);
+                } else if (state == SocketState.OPEN) {
+                    // In keep-alive but between requests. OK to recycle
+                    // processor. Continue to poll for the next request.
+                    recycledProcessors.offer(processor);
+                    proto.endpoint.getPoller().add(socket);
+                } else if (state == SocketState.SENDFILE) {
+                    // Sendfile in progress. If it fails, the socket will be
+                    // closed. If it works, the socket either be added to the
+                    // poller to await more data or processed if there are any
+                    // pipe-lined requests remaining.
+                    connections.put(socket, processor);
                 } else {
                     recycledProcessors.offer(processor);
                 }
@@ -623,10 +678,12 @@ public class Http11AprProtocol implements ProtocolHandler, MBeanRegistration {
             processor.setSocketBuffer(proto.socketBuffer);
             processor.setMaxSavePostSize(proto.maxSavePostSize);
             processor.setServer(proto.server);
+            processor.setClientCertProvider(proto.getClientCertProvider());
+            processor.setMaxCookieCount(proto.getMaxCookieCount());
             register(processor);
             return processor;
         }
-        
+
         protected void register(Http11AprProcessor processor) {
             if (proto.getDomain() != null) {
                 synchronized (this) {
