@@ -14,28 +14,28 @@
  *  See the License for the specific language governing permissions and
  *  limitations under the License.
  */
-
 package org.apache.coyote.http11;
 
 import java.net.InetAddress;
-import java.net.URLEncoder;
 import java.util.Hashtable;
 import java.util.Iterator;
+import java.util.Map;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.Executor;
 import java.util.concurrent.atomic.AtomicInteger;
+
 import javax.management.MBeanRegistration;
 import javax.management.MBeanServer;
 import javax.management.ObjectName;
 
+import org.apache.coyote.AbstractProtocol;
 import org.apache.coyote.ActionCode;
-import org.apache.coyote.ActionHook;
 import org.apache.coyote.Adapter;
-import org.apache.coyote.ProtocolHandler;
 import org.apache.coyote.RequestGroupInfo;
 import org.apache.coyote.RequestInfo;
 import org.apache.tomcat.util.modeler.Registry;
+import org.apache.tomcat.util.net.AbstractEndpoint;
 import org.apache.tomcat.util.net.NioChannel;
 import org.apache.tomcat.util.net.NioEndpoint;
 import org.apache.tomcat.util.net.NioEndpoint.Handler;
@@ -54,10 +54,10 @@ import org.apache.tomcat.util.res.StringManager;
  * @author Costin Manolache
  * @author Filip Hanik
  */
-public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
+public class Http11NioProtocol extends AbstractProtocol implements MBeanRegistration
 {
     protected JSSEImplementation sslImplementation = null;
-    
+
     public Http11NioProtocol() {
         cHandler = new Http11ConnectionHandler( this );
         setSoLinger(Constants.DEFAULT_CONNECTION_LINGER);
@@ -87,7 +87,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
         return attributes.get(key);
     }
 
-    public Iterator getAttributeNames() {
+    public Iterator<String> getAttributeNames() {
         return attributes.keySet().iterator();
     }
 
@@ -101,7 +101,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
         } else {
             return ep.setProperty(name,value); //make sure we at least try to set all properties
         }
-        
+
     }
 
     /**
@@ -127,11 +127,11 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public void init() throws Exception {
         ep.setName(getName());
         ep.setHandler(cHandler);
-        
+
         //todo, determine if we even need these
         ep.getSocketProperties().setRxBufSize(Math.max(ep.getSocketProperties().getRxBufSize(),getMaxHttpHeaderSize()));
         ep.getSocketProperties().setTxBufSize(Math.max(ep.getSocketProperties().getTxBufSize(),getMaxHttpHeaderSize()));
-        
+
         try {
             ep.init();
             sslImplementation = new JSSEImplementation();
@@ -207,9 +207,15 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
 
     // -------------------- Properties--------------------
     protected NioEndpoint ep=new NioEndpoint();
+
+    @Override
+    protected final AbstractEndpoint getEndpoint() {
+        return ep;
+    }
+
     protected boolean secure = false;
 
-    protected Hashtable attributes = new Hashtable();
+    protected Hashtable<String,Object> attributes = new Hashtable<String,Object>();
 
     private int maxKeepAliveRequests=100; // as in Apache HTTPD server
     private int timeout = 300000;   // 5 minutes as in Apache HTTPD server
@@ -219,7 +225,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     private int socketCloseDelay=-1;
     private boolean disableUploadTimeout = true;
     private int socketBuffer = 9000;
-    
+
     private Adapter adapter;
     private Http11ConnectionHandler cHandler;
 
@@ -239,15 +245,15 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public void setPollerThreadCount(int count) {
         ep.setPollerThreadCount(count);
     }
-    
+
     public int getPollerThreadCount() {
         return ep.getPollerThreadCount();
     }
-    
+
     public void setSelectorTimeout(long timeout) {
         ep.setSelectorTimeout(timeout);
     }
-    
+
     public long getSelectorTimeout() {
         return ep.getSelectorTimeout();
     }
@@ -260,7 +266,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public void setExecutor(Executor executor) {
         ep.setExecutor(executor);
     }
-    
+
     public void setUseExecutor(boolean useexec) {
         ep.setUseExecutor(useexec);
     }
@@ -278,7 +284,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
       ep.setThreadPriority(threadPriority);
       setAttribute("threadPriority", "" + threadPriority);
     }
-    
+
     public void setAcceptorThreadPriority(int threadPriority) {
       ep.setAcceptorThreadPriority(threadPriority);
       setAttribute("acceptorThreadPriority", "" + threadPriority);
@@ -296,12 +302,12 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public int getAcceptorThreadPriority() {
       return ep.getAcceptorThreadPriority();
     }
-    
+
     public int getPollerThreadPriority() {
       return ep.getThreadPriority();
     }
-    
-    
+
+
     public boolean getUseSendfile() {
         return ep.getUseSendfile();
     }
@@ -309,7 +315,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public void setUseSendfile(boolean useSendfile) {
         ep.setUseSendfile(useSendfile);
     }
-    
+
     // -------------------- Tcp setup --------------------
 
     public int getBacklog() {
@@ -340,14 +346,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     }
 
     public String getName() {
-        String encodedAddr = "";
-        if (getAddress() != null) {
-            encodedAddr = "" + getAddress();
-            if (encodedAddr.startsWith("/"))
-                encodedAddr = encodedAddr.substring(1);
-            encodedAddr = URLEncoder.encode(encodedAddr) + "-";
-        }
-        return ("http-" + encodedAddr + ep.getPort());
+        return createName("http", getAddress(), ep.getPort());
     }
 
     public boolean getTcpNoDelay() {
@@ -534,7 +533,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
 
     public void setOomParachute(int oomParachute) {
         ep.setOomParachute(oomParachute);
-        setAttribute("oomParachute",oomParachute);
+        setAttribute("oomParachute", Integer.valueOf(oomParachute));
     }
 
     // --------------------  SSL related properties --------------------
@@ -543,19 +542,19 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public void setKeystoreFile(String s ) { ep.setKeystoreFile(s);}
     public void setKeystore(String s) { setKeystoreFile(s);}
     public String getKeystore(){ return getKeystoreFile();}
-    
+
     public String getKeyAlias() { return ep.getKeyAlias();}
     public void setKeyAlias(String s ) { ep.setKeyAlias(s);}
 
-    
+
     public String getAlgorithm() { return ep.getAlgorithm();}
     public void setAlgorithm(String s ) { ep.setAlgorithm(s);}
-    
+
     public void setClientauth(String s) {setClientAuth(s);}
     public String getClientauth(){ return getClientAuth();}
     public String getClientAuth() { return ep.getClientAuth();}
     public void setClientAuth(String s ) { ep.setClientAuth(s);}
-    
+
     public String getKeystorePass() { return ep.getKeystorePass();}
     public void setKeystorePass(String s ) { ep.setKeystorePass(s);}
     public void setKeypass(String s) { setKeystorePass(s);}
@@ -571,18 +570,40 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
     public String getTruststorePass(){return ep.getTruststorePass();}
     public void setTruststoreType(String t){ep.setTruststoreType(t);}
     public String getTruststoreType(){ return ep.getTruststoreType();}
-    
-    
+
+
     public String getSslProtocol() { return ep.getSslProtocol();}
     public void setSslProtocol(String s) { ep.setSslProtocol(s);}
-    
+
     public String getCiphers() { return ep.getCiphers();}
     public void setCiphers(String s) { ep.setCiphers(s);}
-    
+
     public boolean getSSLEnabled() { return ep.isSSLEnabled(); }
     public void setSSLEnabled(boolean SSLEnabled) { ep.setSSLEnabled(SSLEnabled); }
-    
-    
+
+    // Alias for sslEnabledProtocols
+    public void setProtocols(String k) {
+        setSslEnabledProtocols(k);
+    }
+
+    public void setSslEnabledProtocols(String k) {
+        ep.setSslEnabledProtocols(k);
+    }
+
+
+    /**
+     * When client certificate information is presented in a form other than
+     * instances of {@link java.security.cert.X509Certificate} it needs to be
+     * converted before it can be used and this property controls which JSSE
+     * provider is used to perform the conversion. For example it is used with
+     * the AJP connectors, the HTTP APR connector and with the
+     * {@link org.apache.catalina.valves.SSLValve}. If not specified, the
+     * default provider will be used.
+     */
+    protected String clientCertProvider = null;
+    public String getClientCertProvider() { return clientCertProvider; }
+    public void setClientCertProvider(String s) { this.clientCertProvider = s; }
+
 
     // --------------------  Connection handler --------------------
 
@@ -592,10 +613,13 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
         protected static int count = 0;
         protected RequestGroupInfo global = new RequestGroupInfo();
 
-        protected ConcurrentHashMap<NioChannel, Http11NioProcessor> connections =
-            new ConcurrentHashMap<NioChannel, Http11NioProcessor>();
-        protected ConcurrentLinkedQueue<Http11NioProcessor> recycledProcessors = new ConcurrentLinkedQueue<Http11NioProcessor>() {
+        protected final Map<NioChannel, Http11NioProcessor> connections =
+                new ConcurrentHashMap<NioChannel, Http11NioProcessor>();
+        protected ConcurrentLinkedQueue<Http11NioProcessor> recycledProcessors =
+                new ConcurrentLinkedQueue<Http11NioProcessor>() {
+            private static final long serialVersionUID = 1L;
             protected AtomicInteger size = new AtomicInteger(0);
+            @Override
             public boolean offer(Http11NioProcessor processor) {
                 boolean offer = proto.processorCache==-1?true:size.get() < proto.processorCache;
                 //avoid over growing our cache or add after we have stopped
@@ -609,7 +633,8 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                 if (!result) deregister(processor);
                 return result;
             }
-            
+
+            @Override
             public Http11NioProcessor poll() {
                 Http11NioProcessor result = super.poll();
                 if ( result != null ) {
@@ -617,7 +642,8 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                 }
                 return result;
             }
-            
+
+            @Override
             public void clear() {
                 Http11NioProcessor next = poll();
                 while ( next != null ) {
@@ -632,11 +658,15 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
         Http11ConnectionHandler(Http11NioProtocol proto) {
             this.proto = proto;
         }
-        
+
         public void releaseCaches() {
             recycledProcessors.clear();
         }
-        
+
+        /**
+         * Use this only if the processor is not available, otherwise use
+         * {@link #release(NioChannel, Http11NioProcessor).
+         */
         public void release(NioChannel socket) {
             Http11NioProcessor result = connections.remove(socket);
             if ( result != null ) {
@@ -645,10 +675,18 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
             }
         }
 
+
+        public void release(NioChannel socket, Http11NioProcessor processor) {
+            connections.remove(socket);
+            processor.recycle();
+            recycledProcessors.offer(processor);
+        }
+
+
         public SocketState event(NioChannel socket, SocketStatus status) {
             Http11NioProcessor result = connections.get(socket);
 
-            SocketState state = SocketState.CLOSED; 
+            SocketState state = SocketState.CLOSED;
             if (result != null) {
                 if (log.isDebugEnabled()) log.debug("Http11NioProcessor.error="+result.error);
                 // Call the appropriate event
@@ -676,8 +714,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                         (sm.getString("http11protocol.proto.error"), e);
                 } finally {
                     if (state != SocketState.LONG) {
-                        connections.remove(socket);
-                        recycledProcessors.offer(result);
+                        release(socket, result);
                         if (state == SocketState.OPEN) {
                             socket.getPoller().add(socket);
                         }
@@ -696,7 +733,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
             Http11NioProcessor processor = null;
             try {
                 processor = connections.remove(socket);
-                
+
                 if (processor == null) {
                     processor = recycledProcessors.poll();
                 }
@@ -704,12 +741,9 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                     processor = createProcessor();
                 }
 
-                if (processor instanceof ActionHook) {
-                    ((ActionHook) processor).action(ActionCode.ACTION_START, null);
-                }
-                
-                
-                if (proto.ep.getSecure() && (proto.sslImplementation != null)) {
+                processor.action(ActionCode.ACTION_START, null);
+
+                if (proto.ep.isSSLEnabled() && (proto.sslImplementation != null)) {
                     if (socket instanceof SecureNioChannel) {
                         SecureNioChannel ch = (SecureNioChannel)socket;
                         processor.setSslSupport(proto.sslImplementation.getSSLSupport(ch.getSslEngine().getSession()));
@@ -721,19 +755,24 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
 
                 SocketState state = processor.process(socket);
                 if (state == SocketState.LONG) {
-                    // Associate the connection with the processor. The next request 
-                    // processed by this thread will use either a new or a recycled
-                    // processor.
-                    //if (log.isDebugEnabled()) log.debug("Not recycling ["+processor+"] Comet="+((NioEndpoint.KeyAttachment)socket.getAttachment(false)).getComet());
+                    // In the middle of processing a request/response. Keep the
+                    // socket associated with the processor.
+                    connections.put(socket, processor);
+                    socket.getPoller().add(socket);
+                } else if (state == SocketState.OPEN) {
+                    // In keep-alive but between requests. OK to recycle
+                    // processor. Continue to poll for the next request.
+                    release(socket, processor);
+                    socket.getPoller().add(socket);
+                } else if (state == SocketState.SENDFILE) {
+                    // Sendfile in progress. If it fails, the socket will be
+                    // closed. If it works, the socket either be added to the
+                    // poller to await more data or processed if there are any
+                    // pipe-lined requests remaining.
                     connections.put(socket, processor);
-                    if (processor.comet) {
-                        NioEndpoint.KeyAttachment att = (NioEndpoint.KeyAttachment)socket.getAttachment(false);
-                        socket.getPoller().add(socket,att.getCometOps());
-                    } else {
-                        socket.getPoller().add(socket);
-                    }
                 } else {
-                    recycledProcessors.offer(processor);
+                    // Connection closed. OK to recycle the processor.
+                    release(socket, processor);
                 }
                 return state;
 
@@ -758,14 +797,14 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                 Http11NioProtocol.log.error
                     (sm.getString("http11protocol.proto.error"), e);
             }
-            recycledProcessors.offer(processor);
+            release(socket, processor);
             return SocketState.CLOSED;
         }
 
         public Http11NioProcessor createProcessor() {
             Http11NioProcessor processor = new Http11NioProcessor(
               proto.ep.getSocketProperties().getRxBufSize(),
-              proto.ep.getSocketProperties().getTxBufSize(), 
+              proto.ep.getSocketProperties().getTxBufSize(),
               proto.maxHttpHeaderSize,
               proto.ep);
             processor.setAdapter(proto.adapter);
@@ -780,6 +819,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
             processor.setSocketBuffer(proto.socketBuffer);
             processor.setMaxSavePostSize(proto.maxSavePostSize);
             processor.setServer(proto.server);
+            processor.setMaxCookieCount(proto.getMaxCookieCount());
             register(processor);
             return processor;
         }
@@ -803,7 +843,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
                 }
             }
         }
-    
+
         public void deregister(Http11NioProcessor processor) {
             if (proto.getDomain() != null) {
                 synchronized (this) {
@@ -823,7 +863,7 @@ public class Http11NioProtocol implements ProtocolHandler, MBeanRegistration
         }
 
     }
-    
+
 
 
     protected static org.apache.juli.logging.Log log
