@@ -21,8 +21,6 @@ import java.io.IOException;
 import java.io.InterruptedIOException;
 import java.net.InetAddress;
 import java.net.Socket;
-import java.security.AccessController;
-import java.security.PrivilegedAction;
 import java.util.StringTokenizer;
 import java.util.regex.Pattern;
 import java.util.regex.PatternSyntaxException;
@@ -73,16 +71,23 @@ public class Http11Processor implements ActionHook {
     protected static StringManager sm =
         StringManager.getManager(Constants.Package);
 
-    protected static boolean isSecurityEnabled = 
+    protected static boolean isSecurityEnabled =
 	org.apache.coyote.Constants.IS_SECURITY_ENABLED;
 
+    /*
+     * Tracks how many internal filters are in the filter library so they
+     * are skipped when looking for pluggable filters.
+     */
+    private int pluggableFilterIndex = Integer.MAX_VALUE;
+
+
     // ------------------------------------------------------------ Constructor
 
 
     public Http11Processor(int headerBufferSize, JIoEndpoint endpoint) {
 
         this.endpoint = endpoint;
-        
+
         request = new Request();
         inputBuffer = new InternalInputBuffer(request, headerBufferSize);
         request.setInputBuffer(inputBuffer);
@@ -96,7 +101,7 @@ public class Http11Processor implements ActionHook {
         initializeFilters();
 
         // Cause loading of HexUtils
-        int foo = HexUtils.DEC[0];
+        HexUtils.getDec('0');
 
     }
 
@@ -200,7 +205,10 @@ public class Http11Processor implements ActionHook {
      */
     protected SSLSupport sslSupport;
 
-
+    
+    private int maxCookieCount = 200;
+    
+    
     /**
      * Socket associated with the current connection.
      */
@@ -463,7 +471,7 @@ public class Http11Processor implements ActionHook {
      */
     protected void addFilter(String className) {
         try {
-            Class clazz = Class.forName(className);
+            Class<?> clazz = Class.forName(className);
             Object obj = clazz.newInstance();
             if (obj instanceof InputFilter) {
                 inputBuffer.addFilter((InputFilter) obj);
@@ -522,22 +530,6 @@ public class Http11Processor implements ActionHook {
     }
 
 
-    /**
-     * General use method
-     *
-     * @param sArray the StringArray
-     * @param value string
-     */
-    private boolean inStringArray(String sArray[], String value) {
-        for (int i = 0; i < sArray.length; i++) {
-            if (sArray[i].equals(value)) {
-                return true;
-            }
-        }
-        return false;
-    }
-
-
     /**
      * Checks if any entry in the string array starts with the specified value
      *
@@ -670,6 +662,16 @@ public class Http11Processor implements ActionHook {
     }
 
 
+    public int getMaxCookieCount() {
+        return maxCookieCount;
+    }
+
+
+    public void setMaxCookieCount(int maxCookieCount) {
+        this.maxCookieCount = maxCookieCount;
+    }
+
+        
     /**
      * Set the flag to control upload time-outs.
      */
@@ -742,15 +744,14 @@ public class Http11Processor implements ActionHook {
     }
 
     /**
-     * Process pipelined HTTP requests using the specified input and output
-     * streams.
+     * Process pipelined HTTP requests on the specified socket.
+     *
+     * @param theSocket Socket from which the HTTP requests will be read
+     *               and the HTTP responses will be written.
      *
-     * @param input stream from which the HTTP requests will be read
-     * @param output stream which will be used to output the HTTP
-     * responses
      * @throws IOException error during an I/O operation
      */
-    public void process(Socket socket)
+    public void process(Socket theSocket)
         throws IOException {
         RequestInfo rp = request.getRequestProcessor();
         rp.setStage(org.apache.coyote.Constants.STAGE_PARSE);
@@ -764,7 +765,7 @@ public class Http11Processor implements ActionHook {
         localPort = -1;
 
         // Setting up the I/O
-        this.socket = socket;
+        this.socket = theSocket;
         inputBuffer.setInputStream(socket.getInputStream());
         outputBuffer.setOutputStream(socket.getOutputStream());
 
@@ -773,27 +774,29 @@ public class Http11Processor implements ActionHook {
         keepAlive = true;
 
         int keepAliveLeft = maxKeepAliveRequests;
-        int soTimeout = socket.getSoTimeout();
-        int oldSoTimeout = soTimeout;
+        int soTimeout = endpoint.getSoTimeout();
 
-        int threadRatio = (endpoint.getCurrentThreadsBusy() * 100)
-                / endpoint.getMaxThreads();
-        if (threadRatio > 75) {
-            keepAliveLeft = 1;
-        }
-        
-        if (soTimeout != oldSoTimeout) {
-            try {
-                socket.setSoTimeout(soTimeout);
-            } catch (Throwable t) {
-                log.debug(sm.getString("http11processor.socket.timeout"), t);
-                error = true;
+        // When using an executor, these values may return non-positive values
+        int curThreads = endpoint.getCurrentThreadsBusy();
+        int maxThreads = endpoint.getMaxThreads();
+        if (curThreads > 0 && maxThreads > 0) {
+            // Only auto-disable keep-alive if the current thread usage % can be
+            // calculated correctly
+            if ((curThreads*100)/maxThreads > 75) {
+                keepAliveLeft = 1;
             }
         }
 
+        try {
+            socket.setSoTimeout(soTimeout);
+        } catch (Throwable t) {
+            log.debug(sm.getString("http11processor.socket.timeout"), t);
+            error = true;
+        }
+
         boolean keptAlive = false;
 
-        while (started && !error && keepAlive) {
+        while (started && !error && keepAlive && !endpoint.isPaused()) {
 
             // Parsing the request header
             try {
@@ -808,9 +811,14 @@ public class Http11Processor implements ActionHook {
                 inputBuffer.parseRequestLine();
                 request.setStartTime(System.currentTimeMillis());
                 keptAlive = true;
-                if (!disableUploadTimeout) {
+                if (disableUploadTimeout) {
+                    socket.setSoTimeout(soTimeout);
+                } else {
                     socket.setSoTimeout(timeout);
                 }
+                // Set this every time in case limit has been changed via JMX
+                request.getMimeHeaders().setLimit(endpoint.getMaxHeaderCount());
+                request.getCookies().setLimit(getMaxCookieCount());
                 inputBuffer.parseHeaders();
             } catch (IOException e) {
                 error = true;
@@ -821,6 +829,7 @@ public class Http11Processor implements ActionHook {
                 }
                 // 400 - Bad Request
                 response.setStatus(400);
+                adapter.log(request, response, 0);
                 error = true;
             }
 
@@ -835,6 +844,7 @@ public class Http11Processor implements ActionHook {
                     }
                     // 400 - Internal Server Error
                     response.setStatus(400);
+                    adapter.log(request, response, 0);
                     error = true;
                 }
             }
@@ -863,6 +873,7 @@ public class Http11Processor implements ActionHook {
                     log.error(sm.getString("http11processor.request.process"), t);
                     // 500 - Internal Server Error
                     response.setStatus(500);
+                    adapter.log(request, response, 0);
                     error = true;
                 }
             }
@@ -871,7 +882,7 @@ public class Http11Processor implements ActionHook {
             try {
                 rp.setStage(org.apache.coyote.Constants.STAGE_ENDINPUT);
                 // If we know we are closing the connection, don't drain input.
-                // This way uploading a 100GB file doesn't tie up the thread 
+                // This way uploading a 100GB file doesn't tie up the thread
                 // if the servlet has rejected it.
                 if(error)
                     inputBuffer.setSwallowInput(false);
@@ -882,6 +893,7 @@ public class Http11Processor implements ActionHook {
                 log.error(sm.getString("http11processor.request.finish"), t);
                 // 500 - Internal Server Error
                 response.setStatus(500);
+                // No access logging since after service method
                 error = true;
             }
             try {
@@ -1116,7 +1128,7 @@ public class Http11Processor implements ActionHook {
             }
         } else if (actionCode == ActionCode.ACTION_REQ_SET_BODY_REPLAY) {
             ByteChunk body = (ByteChunk) param;
-            
+
             InputFilter savedBody = new SavedRequestInputFilter(body);
             savedBody.setRequest(request);
 
@@ -1306,10 +1318,20 @@ public class Http11Processor implements ActionHook {
 
         // Parse content-length header
         long contentLength = request.getContentLengthLong();
-        if (contentLength >= 0 && !contentDelimitation) {
-            inputBuffer.addActiveFilter
-                (inputFilters[Constants.IDENTITY_FILTER]);
-            contentDelimitation = true;
+        if (contentLength >= 0) {
+            if (contentDelimitation) {
+                // contentDelimitation being true at this point indicates that
+                // chunked encoding is being used but chunked encoding should
+                // not be used with a content length. RFC 2616, section 4.4,
+                // bullet 3 states Content-Length must be ignored in this case -
+                // so remove it.
+                headers.removeHeader("content-length");
+                request.setContentLength(-1);
+            } else {
+                inputBuffer.addActiveFilter
+                        (inputFilters[Constants.IDENTITY_FILTER]);
+                contentDelimitation = true;
+            }
         }
 
         MessageBytes valueMB = headers.getValue("host");
@@ -1328,7 +1350,7 @@ public class Http11Processor implements ActionHook {
         parseHost(valueMB);
 
         if (!contentDelimitation) {
-            // If there's no content length 
+            // If there's no content length
             // (broken HTTP/1.0 or HTTP/1.1), assume
             // the client is not broken and didn't send a body
             inputBuffer.addActiveFilter
@@ -1336,6 +1358,9 @@ public class Http11Processor implements ActionHook {
             contentDelimitation = true;
         }
 
+        if (error) {
+            adapter.log(request, response, 0);
+        }
     }
 
 
@@ -1396,7 +1421,7 @@ public class Http11Processor implements ActionHook {
             int port = 0;
             int mult = 1;
             for (int i = valueL - 1; i > colonPos; i--) {
-                int charValue = HexUtils.DEC[(int) valueB[i + valueS]];
+                int charValue = HexUtils.getDec(valueB[i + valueS]);
                 if (charValue == -1) {
                     // Invalid character
                     error = true;
@@ -1415,16 +1440,42 @@ public class Http11Processor implements ActionHook {
 
 
     /**
-     * Check for compression
+     * Check if the resource could be compressed, if the client supports it.
      */
     private boolean isCompressable() {
 
-        // Nope Compression could works in HTTP 1.0 also
-        // cf: mod_deflate
+        // Check if content is not already gzipped
+        MessageBytes contentEncodingMB =
+            response.getMimeHeaders().getValue("Content-Encoding");
 
-        // Compression only since HTTP 1.1
-        // if (! http11)
-        //    return false;
+        if ((contentEncodingMB != null)
+            && (contentEncodingMB.indexOf("gzip") != -1))
+            return false;
+
+        // If force mode, always compress (test purposes only)
+        if (compressionLevel == 2)
+           return true;
+
+        // Check if sufficient length to trigger the compression
+        long contentLength = response.getContentLengthLong();
+        if ((contentLength == -1)
+            || (contentLength > compressionMinSize)) {
+            // Check for compatible MIME-TYPE
+            if (compressableMimeTypes != null) {
+                return (startsWithStringArray(compressableMimeTypes,
+                                              response.getContentType()));
+            }
+        }
+
+        return false;
+    }
+
+
+    /**
+     * Check if compression should be used for this resource. Already checked
+     * that the resource could be compressed if the client supports it.
+     */
+    private boolean useCompression() {
 
         // Check if browser support gzip encoding
         MessageBytes acceptEncodingMB =
@@ -1434,15 +1485,7 @@ public class Http11Processor implements ActionHook {
             || (acceptEncodingMB.indexOf("gzip") == -1))
             return false;
 
-        // Check if content is not allready gzipped
-        MessageBytes contentEncodingMB =
-            response.getMimeHeaders().getValue("Content-Encoding");
-
-        if ((contentEncodingMB != null)
-            && (contentEncodingMB.indexOf("gzip") != -1))
-            return false;
-
-        // If force mode, allways compress (test purposes only)
+        // If force mode, always compress (test purposes only)
         if (compressionLevel == 2)
            return true;
 
@@ -1460,18 +1503,7 @@ public class Http11Processor implements ActionHook {
             }
         }
 
-        // Check if suffisant len to trig the compression
-        long contentLength = response.getContentLengthLong();
-        if ((contentLength == -1)
-            || (contentLength > compressionMinSize)) {
-            // Check for compatible MIME-TYPE
-            if (compressableMimeTypes != null) {
-                return (startsWithStringArray(compressableMimeTypes,
-                                              response.getContentType()));
-            }
-        }
-
-        return false;
+        return true;
     }
 
 
@@ -1512,10 +1544,13 @@ public class Http11Processor implements ActionHook {
         }
 
         // Check for compression
+        boolean isCompressable = false;
         boolean useCompression = false;
         if (entityBody && (compressionLevel > 0)) {
-            useCompression = isCompressable();
-
+            isCompressable = isCompressable();
+            if (isCompressable) {
+                useCompression = useCompression();
+            }
             // Change content-length to -1 to force chunking
             if (useCompression) {
                 response.setContentLength(-1);
@@ -1544,7 +1579,7 @@ public class Http11Processor implements ActionHook {
                 (outputFilters[Constants.IDENTITY_FILTER]);
             contentDelimitation = true;
         } else {
-            if (entityBody && http11 && keepAlive) {
+            if (entityBody && http11) {
                 outputBuffer.addActiveFilter
                     (outputFilters[Constants.CHUNKED_FILTER]);
                 contentDelimitation = true;
@@ -1558,24 +1593,29 @@ public class Http11Processor implements ActionHook {
         if (useCompression) {
             outputBuffer.addActiveFilter(outputFilters[Constants.GZIP_FILTER]);
             headers.setValue("Content-Encoding").setString("gzip");
+        }
+        // If it might be compressed, set the Vary header
+        if (isCompressable) {
             // Make Proxies happy via Vary (from mod_deflate)
-            headers.setValue("Vary").setString("Accept-Encoding");
+            MessageBytes vary = headers.getValue("Vary");
+            if (vary == null) {
+                // Add a new Vary header
+                headers.setValue("Vary").setString("Accept-Encoding");
+            } else if (vary.equals("*")) {
+                // No action required
+            } else {
+                // Merge into current header
+                headers.setValue("Vary").setString(
+                        vary.getString() + ",Accept-Encoding");
+            }
         }
 
-        // Add date header
-        String date = null;
-        if (isSecurityEnabled){
-            date = (String)AccessController.doPrivileged(
-                    new PrivilegedAction() {
-                        public Object run(){
-                            return FastHttpDateFormat.getCurrentDate();
-                        }
-                    }
-            );
-        } else {
-            date = FastHttpDateFormat.getCurrentDate();
+        // Add date header unless application has already set one (e.g. in a
+        // Caching Filter)
+        if (headers.getValue("Date") == null) {
+            headers.setValue("Date").setString(
+                    FastHttpDateFormat.getCurrentDate());
         }
-        headers.setValue("Date").setString(date);
 
         // FIXME: Add transfer encoding header
 
@@ -1599,8 +1639,10 @@ public class Http11Processor implements ActionHook {
 
         // Add server header
         if (server != null) {
+            // Always overrides anything the app might set
             headers.setValue("Server").setString(server);
-        } else {
+        } else if (headers.getValue("Server") == null) {
+            // If app didn't set the header, use the default
             outputBuffer.write(Constants.SERVER_BYTES);
         }
 
@@ -1637,6 +1679,8 @@ public class Http11Processor implements ActionHook {
         //inputBuffer.addFilter(new GzipInputFilter());
         outputBuffer.addFilter(new GzipOutputFilter());
 
+        pluggableFilterIndex = inputBuffer.filterLibrary.length;
+
     }
 
 
@@ -1655,7 +1699,7 @@ public class Http11Processor implements ActionHook {
                 (inputFilters[Constants.CHUNKED_FILTER]);
             contentDelimitation = true;
         } else {
-            for (int i = 2; i < inputFilters.length; i++) {
+            for (int i = pluggableFilterIndex; i < inputFilters.length; i++) {
                 if (inputFilters[i].getEncodingName()
                     .toString().equals(encodingName)) {
                     inputBuffer.addActiveFilter(inputFilters[i]);
